package de.osci.osci12.samples.impl.crypto;

import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.IOException;
import java.io.InputStream;

import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import java.util.Enumeration;


/**
 * Diese Klasse ist eine Beispiel-Implementierung der abstrakten Signer-Klasse.
 * Fr die Verwendung wird ein PKCS#12-Keystore in Form einer *.p12-Datei bentigt.
 * Die Implementierung ist fr Testzwecke bestimmt, sie greift auf den ersten
 * verfgbaren Alias zu. Die PIN fr dessen Privatschlssel muss die gleiche sein
 * wie die des Keystores.
 *
 * Diese einfache Implementierung hlt die PIN des Keystores als Character-Array
 * im Arbeitsspeicher, sie wird als String bergeben. Es wird Anwendern empfohlen,
 * eigene Implementierungen zu schreiben, die die PIN in der Methode sign(...)
 * abfragen und nach Gebrauch wieder lschen oder anderweitig fr ein sicheres
 * Pin-Cashing zu sorgen.
 *
 * <p>Copyright  2008 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.3
 * @see de.osci.osci12.extinterfaces.crypto.Signer
 */
public class PKCS12Signer extends de.osci.osci12.extinterfaces.crypto.Signer
{
  private static Log log = LogFactory.getLog(PKCS12Signer.class);
  private X509Certificate cert;
  private String alias;
  private char[] pin = null;
  private KeyStore keyStore = null;

  /**
   * Creates a new PKCS12Signer object.
   *
   * @param p12_fileName undocumented
   * @param pin undocumented
   *
   * @throws KeyStoreException undocumented
   * @throws CertificateException undocumented
   * @throws NoSuchAlgorithmException undocumented
   * @throws IOException undocumented
   */
  public PKCS12Signer(String p12_fileName, String pin)
               throws KeyStoreException, 
                      CertificateException, 
                      NoSuchAlgorithmException, 
                      IOException
  {
    init(getClass().getResourceAsStream(p12_fileName), pin);
  }

  /**
   * Creates a new PKCS12Signer object.
   *
   * @param in undocumented
   * @param pin undocumented
   *
   * @throws KeyStoreException undocumented
   * @throws CertificateException undocumented
   * @throws NoSuchAlgorithmException undocumented
   * @throws IOException undocumented
   */
  public PKCS12Signer(InputStream in, String pin)
               throws KeyStoreException, 
                      CertificateException, 
                      NoSuchAlgorithmException, 
                      IOException
  {
    init(in, pin);
  }

  private void init(InputStream in, String pin)
             throws KeyStoreException, 
                    CertificateException, 
                    NoSuchAlgorithmException, 
                    IOException
  {
    this.pin = pin.toCharArray();

    if (DialogHandler.getSecurityProvider() == null)
      keyStore = java.security.KeyStore.getInstance("PKCS12");
    else
    {
      keyStore = java.security.KeyStore.getInstance("PKCS12", DialogHandler.getSecurityProvider());
    }

    keyStore.load(in, this.pin);

    String al = null;
    Enumeration e = keyStore.aliases();

    while (e.hasMoreElements())

      if (keyStore.isKeyEntry(al = (String) e.nextElement()))
        break;

    if (al == null)
      throw new NullPointerException("No private key found in keystore.");

    this.alias = al;
    cert = (X509Certificate) keyStore.getCertificate(alias);
  }

  /**
   * Sollte den Namen des Herstellers zurckgeben.
   * @return Herstellername
   */
  public String getVendor()
  {
    return "BOS";
  }

  /**
   * Liefert die Versionsnummer
   * @return Versionsnummer
   */
  public String getVersion()
  {
    return "1.3";
  }

  /**
   * Liefert den im {@link DialogHandler} gesetzten Default-Signaturalgorithmus.
   * @return Signaturalgorithmus-Identifier
   * @see DialogHandler#getSignatureAlgorithm()
   */
  public String getAlgorithm()
  {
	  return DialogHandler.getSignatureAlgorithm();
  }
  
  /**
   *  Liefert das Signaturzertifikat.
   *
   *@return  das Signaturzertifikat
   */
  public X509Certificate getCertificate()
  {
    return cert;
  }

  /**
   *  Signiert das bergebene Byte-Array
   *
   *@param  data die zu signierenden Daten
   * @param algorithm Signaturalgorithmus
   * @return die Signatur
   * @throws SignatureException
   */
  public byte[] sign(byte[] data, String algorithm) throws SignatureException
  {
    try
    {
      java.security.Signature sigengine;
      if (log.isDebugEnabled())
    	  log.debug("Algorithm: " + algorithm);

      if (DialogHandler.getSecurityProvider() == null)
        sigengine = java.security.Signature.getInstance((String) Constants.JCA_JCE_MAP.get(algorithm));
      else
        sigengine = java.security.Signature.getInstance((String) Constants.JCA_JCE_MAP.get(algorithm),
                                                        DialogHandler.getSecurityProvider());

      sigengine.initSign(((java.security.PrivateKey) keyStore.getKey(alias, pin)));
      sigengine.update(data);

      if (log.isDebugEnabled())
        log.debug("Fertig mit dem Signieren");

      return sigengine.sign();
    }
    catch (Exception ex)
    {
      throw new SignatureException(DialogHandler.text.getString("signature_creation_error") + " - " + ex.getClass() +
                                   ": " + ex.getLocalizedMessage());
    }
  }
}
