package de.osci.helper;

import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.encryption.Crypto;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import java.security.GeneralSecurityException;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.spec.IvParameterSpec;


/**
 * Diese Erweiterung der Klasse java.io.FilterOutputStream fhrt eine symmetrische
 * Ver-/Entschlsselung der geschriebenen Daten durch. Der Initialisierungsvektor
 * wird gem der XML-Encryption-Spezifikation den Inhaltsdaten hinzugefgt bzw.
 * entnommen.
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 * @see de.osci.helper.SymCipherInputStream
 */
public class SymCipherOutputStream extends FilterOutputStream
{
  //  private static Log log = LogFactory.getLog(SymCipherOutputStream.class);
  private Cipher cipher;
  private int index = 0;
  private byte[] iv;
  private SecretKey symKey;
  private boolean encrypt;
  private byte[] tmp;

  /**
   * Creates a new SymCipherOutputStream object.
   *
   * @param outStream undocumented
   * @param symKey undocumented
   * @param iv undocumented
   *
   * @throws IOException undocumented
   */
  public SymCipherOutputStream(OutputStream outStream, SecretKey symKey, byte[] iv)
                        throws IOException
  {
    this(outStream, symKey, true, iv);
  }

  /**
   * Creates a new SymCipherOutputStream object.
   *
   * @param outStream undocumented
   * @param symKey undocumented
   * @param encrypt undocumented
   *
   * @throws IOException undocumented
   */
  public SymCipherOutputStream(OutputStream outStream, SecretKey symKey, boolean encrypt)
                        throws IOException
  {
    this(outStream, symKey, encrypt, null);
  }

  SymCipherOutputStream(OutputStream outStream, SecretKey symKey, boolean encrypt, byte[] iv)
                 throws IOException
  {
    super(outStream);
    this.symKey = symKey;
    this.encrypt = encrypt;

    try
    {
      if (DialogHandler.getSecurityProvider() == null)
        cipher = Cipher.getInstance((String) Constants.JCA_JCE_MAP.get(Crypto.getCipherAlgoID(symKey)));
      else
        cipher = Cipher.getInstance((String) Constants.JCA_JCE_MAP.get(Crypto.getCipherAlgoID(symKey)),
                                    DialogHandler.getSecurityProvider());

      if (symKey.getAlgorithm().startsWith("DESede"))
          this.iv = new byte[8];
        else
          this.iv = new byte[16];

      if (encrypt)
      {
        if (iv == null)
        	this.iv = Tools.createRawRandom(this.iv.length);
        else
        	System.arraycopy(iv, 0, this.iv, 0, this.iv.length);
        
        IvParameterSpec algoParamSpec = new javax.crypto.spec.IvParameterSpec(this.iv);
        cipher.init(Cipher.ENCRYPT_MODE, symKey, algoParamSpec);

//        this.iv = cipher.getIV();
        outStream.write(this.iv);
      }
    }
    catch (GeneralSecurityException ex)
    {
      if (encrypt)
        throw new IOException(DialogHandler.text.getString("encryption_error"));
      else
        throw new IOException(DialogHandler.text.getString("decryption_error"));
    }
  }

  /**
   * undocumented
   *
   * @param b undocumented
   *
   * @throws IOException undocumented
   */
  public void write(int b) throws IOException
  {
    //    out.write(new byte[] { (byte) b }, 0, 1);
    write(new byte[] { (byte) b }, 0, 1);
  }

  /**
   * undocumented
   *
   * @param b undocumented
   * @param off undocumented
   * @param len undocumented
   *
   * @throws IOException undocumented
   * @throws IndexOutOfBoundsException undocumented
   */
  public void write(byte[] b, int off, int len) throws IOException
  {
    if ((off | len | (b.length - (len + off)) | (off + len)) < 0)
      throw new IndexOutOfBoundsException();

    if ((!encrypt) && (index < iv.length))
    {
      if (len < (iv.length - index))
      {
        for (int i = 0; i < len; i++)
          iv[index++] = b[off + i];
      }
      else
      {
        int i;

        for (i = 0; i < (iv.length - index); i++)
          iv[i + index] = b[off + i];

        index += i;

        try
        {
          IvParameterSpec algoParamSpec = new javax.crypto.spec.IvParameterSpec(iv);
          cipher.init(Cipher.DECRYPT_MODE, symKey, algoParamSpec);
          tmp = cipher.update(b, off + i, len - i);

          if (tmp != null)
            out.write(tmp);
        }
        catch (Exception ex)
        {
          throw new IOException(DialogHandler.text.getString("decryption_error"));
        }
      }
    }
    else
    {
      //      out.write(b, off, len);
      if (b.length > 0)
      {
        tmp = cipher.update(b, off, len);

        if (tmp != null)
          out.write(tmp);
      }
    }
  }

  /**
   * undocumented
   *
   * @throws IOException undocumented
   */
  public void close() throws IOException
  {
    try
    {
      out.write(cipher.doFinal());
    }
    catch (Exception ex)
    {
      if (encrypt)
        throw new IOException(DialogHandler.text.getString("encryption_error"));
      else
        throw new IOException(DialogHandler.text.getString("decryption_error"));
    }

    out.flush();
  }

  /**
   * undocumented
   *
   * @return undocumented
   */
  public byte[] getIv()
  {
    return iv;
  }
}
