package de.osci.osci12.extinterfaces.crypto;

import java.security.SignatureException;
import java.security.cert.X509Certificate;

import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;


/**
 * Diese abstrakte Klasse stellt die Schnittstelle der Bibliothek fr
 * die Anbindung von Signier-Modulen (Crypto-Token) dar. Anwendungen, die
 * OSCI-Nachrichten signieren wollen, mssen an das signierende Rollenobjekt
 * eine Implementation dieser Klasse bergeben.
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 * @see de.osci.osci12.roles.Role#setSignatureAlgorithm(String)
 */
public abstract class Signer
{
  /**
   * Liefert die Versionsnummer.
   * @return Versionsnummer
   */
  public abstract String getVersion();

  /**
   * Sollte den Namen des Herstellers zurckgeben.
   * @return Herstellername
   */
  public abstract String getVendor();

  /**
   * Die Implementierung dieser Methode muss das Signaturzertifikat zurckgeben.
   * @return Signaturzertifikat
   */
  public abstract X509Certificate getCertificate();

  /**
   * Diese Methode sollte den Signaturalgorithmus als XML-Identifier zurckgeben,
   * den die Implementierung bei der Erzeugung einer Signatur verwendet.
   * Gibt diese Methode null zurck, wird der im DialogHandler gesetzte
   * Default-Algorithmus verwendet.
   * @return Signaturalgorithmus
   * @see Constants#SIGNATURE_ALGORITHM_RSA_SHA256
   * @see Constants#SIGNATURE_ALGORITHM_RSA_SHA512
   * @see Constants#SIGNATURE_ALGORITHM_RSA_RIPEMD160
   * @see Constants#SIGNATURE_ALGORITHM_ECDSA_SHA256
   * @see Constants#SIGNATURE_ALGORITHM_ECDSA_SHA512
   * @see DialogHandler#setSignatureAlgorithm(String)
   */
  public abstract String getAlgorithm();

  /**
   * Die Implementierung dieser Methode muss das bergebene Byte-Array signieren.
   * Der algorithm-Parameter hat keine weitere Bedeutung mehr, da ab Version 1.3 die
   * Signer-Implemetierung ber die Methode getAlgorithm() den Algorithmus festlegt.
   * Die OSCI-Bibliothek fragt vor dem Aufruf von sign(byte[], String) den Signaturalgorithmus
   * ab und bergibt den String wieder an die sign-Methode.
   * @param hash zu signierendes Byte-Array
   * @param algorithm Signaturalgorithmus
   * @return Signatur-Wert
   * @throws SignatureException wenn beim Erstellen der Signatur ein Fehler auftritt
   * @throws de.osci.osci12.common.OSCICancelledException wenn der Vorgang vom Anwender abgebrochen wird
   */
  public abstract byte[] sign(byte[] hash, String algorithm)
                       throws SignatureException,
                              de.osci.osci12.common.OSCICancelledException;
}
