package de.osci.osci12.messageparts;

import de.osci.helper.Base64;
import de.osci.helper.Base64OutputStream;
import de.osci.helper.NullOutputStream;
import de.osci.helper.SymCipherInputStream;
import de.osci.helper.SymCipherOutputStream;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.encryption.OSCICipherException;
import de.osci.osci12.extinterfaces.OSCIDataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateFactory;

import java.util.Enumeration;
import java.util.Hashtable;

import javax.crypto.SecretKey;


/**
 * <p>Die Attachment-Klasse reprsentiert einen Anhang einer OSCI-Nachricht.
 * Attachments werden in Content-Elementen mittels eines href-Attributs referenziert.
 * Sie besitzen hierfr einen Identifier (refId), der innerhalb der Nachricht, an die das
 * Attachment gehngt wird, eindeutig sein muss. </p>
 * <p>Ein Attachment kann mit einem eigenen symmetrischen schlssel versehen werden.
 * Hierdurch kann in verschiedenen verschlsselten Inhaltsdatencontainern (EncryptedDataOSCI)
 * dasselbe Attachment referenziert werden.</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 * @see Content
 * @see ContentContainer
 */
public class Attachment extends MessagePart
{
  private static Log log = LogFactory.getLog(Attachment.class);
  private final static int STATE_OF_ATTACHMENT_CLIENT = 0;
  public final static int STATE_OF_ATTACHMENT_PARSING = 1;
  public final static int STATE_OF_ATTACHMENT_ENCRYPTED = 2;
  SecretKey secretKey;
  OSCIDataSource swapBuffer;
  /** Setzt den Status ob das Attachment verschlsselt eingestellt werden soll */
  boolean encrypt;
  private String contentType = "application/octet-stream";
  private Hashtable encryptedDigestValues = new Hashtable();
  int stateOfAttachment = STATE_OF_ATTACHMENT_CLIENT;
  private String boundary_string = DialogHandler.boundary;
  protected boolean base64 = true;
  Hashtable mimeHeaders;

  /**
   * Dieser Konstruktor wird von dem Parser aufgerufen
   *
   * @param ins
   * @param refId
   */
  Attachment(java.io.InputStream ins, String refId, long length, String transportDigestAlgorithm)
      throws IllegalArgumentException, 
             IOException, 
             NoSuchAlgorithmException
  {
    setRefID(refId);
    this.length = length;

    if (ins != null)
    {
      this.setInputStream(ins, false, length, transportDigestAlgorithm);
    }
  }

  /**
   * Erzeugt ein neues Attachment-Objekt aus dem InputStream.
   *
   * @param ins der InputStream, aus dem die Daten gelesen und an die Nachricht
   * angehngt werden
   * @param refId Identifier des Anhangs, z.B. Dateiname
   * @param symmetricCipherAlgorithm symmetrischer Verschlsselungsalgorithmus
   * Die mglichen Werte sind Constants.SYMMETRIC_CIPHER_ALGORITHM_TDES_CBC und
   * Constants.SYMMETRIC_CIPHER_ALGORITHM_AES128, -129, -256
   * @see de.osci.osci12.common.Constants
   *
   */
  public Attachment(java.io.InputStream ins, String refId, String symmetricCipherAlgorithm)
             throws IllegalArgumentException, 
                    IOException, 
                    NoSuchAlgorithmException
  {
    this(ins, refId, de.osci.osci12.encryption.Crypto.createSymKey(symmetricCipherAlgorithm));
  }

  /**
   * Erzeugt ein neues Attachment-Objekt aus dem InputStream. Der geheime
   * schlssel wird fr die Verschlsselung des Attachments benutzt.
   *
   * @param ins der InputStream, aus dem die Daten gelesen und an die Nachricht
   * angehngt werden
   * @param refId Identifier des Anhangs, z.B. Dateiname. Dieser Identifier muss innerhalb
   * der Nachricht, an die das Attachment gehngt wird, eindeutig sein
   * @param secretKey der Secret-Key mit dem verschlsselt werden soll
   * Der bergebene schlssel muss daher fr diesem Algorithmus anwendbar sein
   * Wird dieser Parameter mit 'null' bergeben, wird ein neuer
   * AES-256-schlssel erzeugt
   *
   */
  public Attachment(java.io.InputStream ins, String refId, SecretKey secretKey)
             throws IllegalArgumentException, 
                    IOException, 
                    NoSuchAlgorithmException
  {
    if (ins == null)
    {
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " ins");
    }

    if (refId == null)
    {
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_secondargument") + " refId");
    }

    if (secretKey == null)
    {
      this.secretKey = de.osci.osci12.encryption.Crypto.createSymKey(Constants.SYMMETRIC_CIPHER_ALGORITHM_AES256);
    }
    else
    {
      if (!secretKey.getAlgorithm().equals("DESede") && !secretKey.getAlgorithm().equals("AES"))
        throw new IllegalArgumentException(DialogHandler.text.getString("encryption_algorithm_not_supported"));

      this.secretKey = secretKey;
    }

    this.encrypt = true;
    setRefID(refId);

    if (log.isDebugEnabled())
      log.debug("RefId Name: " + refId);

    this.makeTempFile(ins);
  }

  /**
   * Erzeugt ein neues Attachment-Objekt aus dem InputStream. Das Attachment wird unverschlsselt bertragen.
   *
   * @param ins der InputStream, aus dem die Daten gelesen und an die Nachricht
   * angehngt werden.
   * @param refId Identifier des Anhangs, z.B. Dateiname. Dieser Identifier muss innerhalb
   * der Nachricht, an die das Attachment gehngt wird, eindeutig sein.
   */
  public Attachment(java.io.InputStream ins, String refId)
             throws IllegalArgumentException, 
                    IOException, 
                    NoSuchAlgorithmException
  {
    if (ins == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " null");

    if (refId == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_secondargument") + " null");

    this.encrypt = false;
    setRefID(refId);

    if (log.isDebugEnabled())
      log.debug("RefId Name: " + refId);

    this.makeTempFile(ins);
  }

  /**
   * Setzt das refID-Attribut des Attachments. Da der hier verwendete String
   * auch als Referenz im XML der OSCI-Nachricht verwendet wird, empfiehlt sich
   * die Anwendung von URL-encoding.
   *
   * @param  id  refID-String
   * @see #getRefID()
   * @see java.net.URLEncoder
   */
  public void setRefID(String id)
  {
    super.setRefID(id);
  }

  /**
   *
   * Liefert das refID-Attribut des Attachments. Da fr die verwendeten Strings
   * URL-encoding empfohlen wird, sollte ggf. URL-decoding angewendet werden.
   *
   * @see #setRefID(String)
   * @see java.net.URLDecoder
   */
  public String getRefID()
  {
    return super.getRefID();
  }


  /**
   * Diese Methode liefert die Lnge des Attachments in Byte.
   * @return Lnge des Attachments
   */
  public long getLength()
  {
    if (base64)
      return Base64.calcB64Length(this.length);
    else

      return this.length;
  }

  /**
   * Liefert den InputStream der Daten zuRck, wenn das Attachment einer
   * empfangenen Nachricht entnommen wurde.
   * @return InputStream der Anhangsdaten
   */
  public InputStream getStream() throws OSCICipherException, 
                                        IOException
  {
    if (log.isDebugEnabled())
      log.debug("Attachment Object: ");

    if (swapBuffer == null)
      return null;

    InputStream is = null;
    is = swapBuffer.getInputStream();
    is.reset();

    if (encrypt || (stateOfAttachment == STATE_OF_ATTACHMENT_ENCRYPTED))
    {
      if ((secretKey == null) && (stateOfAttachment != STATE_OF_ATTACHMENT_PARSING))
      {
        throw new OSCICipherException("no_secret_key");
      }

      //      Base64InputStream base64in = new Base64InputStream(is);
      if (log.isDebugEnabled())
        log.debug("Attachment ist Base64 codiert.");

      if (log.isDebugEnabled())
        log.debug("Attachment ist verschlsselt codiert.");

      return new SymCipherInputStream(is, secretKey, false);
    }
    else
    {
      if (log.isDebugEnabled())
        log.debug("Daten sind nur BASE64igt");

      //      return new Base64InputStream(is);
      return is;
    }
  }

  /**
   * Diese Methode liefert den Content der Binrdaten im Mime-Format.
   * @return Content-Type der Daten
   * #setContentType(String)
   */
  public String getContentType()
  {
    return contentType;
  }

  /**
   * Diese Methode setzt den Content-Type (MIME-Content-Type) der Daten
   * @param contentType Content-Type der Daten, sollte ein gltiges
   * Mime-Format sein (z.B. 'text/html' oder 'image/gif')
   * #getContentType()
   */
  public void setContentType(String contentType)
  {
    this.contentType = contentType;
  }

  /**
   * Diese Methode wird nur vom Parser aufgerufen
   * @param ins Inputstream des Attachment
   * @param encrypted Status des Attachments (sind die Daten Verschlsselt)
   * @param length Laenge des Stream
   */
  void setInputStream(InputStream ins, boolean encrypted, long length, String transportDigestAlgorithm)
               throws IOException, 
                      NoSuchAlgorithmException
  {
    if (log.isDebugEnabled())
      log.debug("entry SetStream.");

    this.length = length;
    this.encrypt = encrypted;

    if (ins != null)
    {
		swapBuffer = DialogHandler.getNewDataBuffer();
		OutputStream out = swapBuffer.getOutputStream();
		MessageDigest encMsgDigest = null;

		if (transportDigestAlgorithm != null)
    	{

    		if (DialogHandler.getSecurityProvider() == null)
    			encMsgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(transportDigestAlgorithm));
    		else
    			encMsgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(transportDigestAlgorithm),
                                                 DialogHandler.getSecurityProvider());

    		out = new DigestOutputStream(out, encMsgDigest);
    	}

      java.io.BufferedInputStream bufferdIn = new java.io.BufferedInputStream(ins, 1024);
      int count = 0;
      byte[] bytesIn = new byte[1024];

      while ((count = bufferdIn.read(bytesIn)) > -1)
      {
        out.write(bytesIn, 0, count);
      }

      ins.close();
      out.close();
      bufferdIn.close();
      if (transportDigestAlgorithm != null)
    	  encryptedDigestValues.put(transportDigestAlgorithm, encMsgDigest.digest());
      this.length = swapBuffer.getLength();
    }
    else
    {
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " null");
    }
  }

  /**
   * Diese Methode liest die Daten aus dem InputStream und verschlsselt sie und
   * ermittelt den Digest des Attachments. Das Verschlsselte Ergebnis wird in einen
   * OSCIDataSource geschriben.
   * @throws IOException
   * @throws NoSuchAlgorithmException
   */
  private void makeTempFile(InputStream ins) throws IOException, 
                                                    NoSuchAlgorithmException
  {
    if (swapBuffer != null)
    {
      return;
    }

    if (log.isDebugEnabled())
      log.debug("MakeTempFile wurde aufgerufen.");

    swapBuffer = DialogHandler.getNewDataBuffer();

    OutputStream out = swapBuffer.getOutputStream();
    MessageDigest encMsgDigest = null;
    MessageDigest msgDigest = null;

    if (DialogHandler.getSecurityProvider() == null)
      msgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(DialogHandler.getDigestAlgorithm()));
    else
      msgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(DialogHandler.getDigestAlgorithm()),
                                            DialogHandler.getSecurityProvider());

    if (encrypt)
    {
      if (DialogHandler.getSecurityProvider() == null)
        encMsgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(DialogHandler.getDigestAlgorithm()));
      else
        encMsgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(DialogHandler.getDigestAlgorithm()),
                                                 DialogHandler.getSecurityProvider());

      if (log.isDebugEnabled())
        log.debug("Verschlsseltes Attachment wird erstellt");

      out = new DigestOutputStream(out, encMsgDigest);
      out = new SymCipherOutputStream(out, secretKey, true);
    }
    else if (log.isDebugEnabled())
      log.debug("Unverschlsseltes Attachment wird erstellt");

    out = new DigestOutputStream(out, msgDigest);

    java.io.BufferedInputStream bufferdIn = new java.io.BufferedInputStream(ins, 1024);
    int count = 0;
    byte[] bytesIn = new byte[1024];
    length = 0;

    while ((count = bufferdIn.read(bytesIn)) > -1)
      out.write(bytesIn, 0, count);

    out.flush();
    swapBuffer.getOutputStream().flush();
    out.close();
    bufferdIn.close();
    digestValues.put(DialogHandler.getDigestAlgorithm(), msgDigest.digest());

    if (encrypt)
      encryptedDigestValues.put(DialogHandler.getDigestAlgorithm(), encMsgDigest.digest());
    else
      encryptedDigestValues.put(DialogHandler.getDigestAlgorithm(), digestValues.get(DialogHandler.getDigestAlgorithm()));

    length = swapBuffer.getLength();
  }

  /**
   * undocumented
   *
   * @return undocumented
   */
  public boolean hasDigestValue(String digestAlgorithm)
  {
    return digestValues.get(digestAlgorithm) != null;
  }

  /**
   * Liefert den Hashwert des Attachments.
   *
   * @return Hashwert des Attachments
   */
  public byte[] getDigestValue(String digestAlgorithm) throws IOException, 
                                        NoSuchAlgorithmException
  {
    if (log.isDebugEnabled())
      log.debug("Digest vorhanden? " + digestValues.get(digestAlgorithm) + " stateOfObjekt:" + stateOfAttachment);

    if (digestValues.get(digestAlgorithm) == null)
    {
      digestValues.put(digestAlgorithm, createDigest(digestAlgorithm, encrypt));
    }

    if (log.isDebugEnabled())
      log.debug("Der Digest Value des Attachment ist: " + new String((byte[])digestValues.get(digestAlgorithm)));

    return (byte[])digestValues.get(digestAlgorithm);
  }

  private byte[] createDigest(String digestAlgorithm, boolean enc) throws IOException, 
                                     NoSuchAlgorithmException
  {
    InputStream in = null;
    in = swapBuffer.getInputStream();

    if (in == null)
    {
      throw new IOException(DialogHandler.text.getString("io_exception"));
    }

    in.reset();

    MessageDigest msgDigest;

    if (DialogHandler.getSecurityProvider() == null)
      msgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(digestAlgorithm));
    else
      msgDigest = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(digestAlgorithm),
                                            DialogHandler.getSecurityProvider());

    if (log.isDebugEnabled())
      log.debug("State of Attachment: " + stateOfAttachment + " Verschlsselt: " + encrypt);

    InputStream inStream = in;

    if (!(stateOfAttachment == STATE_OF_ATTACHMENT_PARSING) && enc)
    {
      if (log.isDebugEnabled())
        log.debug("Das Attachment ist verschlsselt.");

      inStream = new SymCipherInputStream(in, secretKey, false);
    }

    DigestOutputStream out = new DigestOutputStream(new NullOutputStream(), msgDigest);
    byte[] bytes = new byte[1024];
    int count = 1024;

    while ((count = inStream.read(bytes)) > -1)
      out.write(bytes, 0, count);

    inStream.close();
    out.close();
    return msgDigest.digest();
  }

  /**
   * Setzt den MIME-Trennstring. Wird von Anwendungen normalerweise nicht aufgerufen.
   * @param boundary Trenner
   */
  public void setBoundary(String boundary)
  {
    boundary_string = boundary;
  }

  /**
   * Setzt das Encoding dea MIME-Boundary-Abschnitts auf Base64.
   * @param b64 true -> Daten werden Base64 codiert.
   */
  public void setBase64Encoding(boolean b64)
  {
    base64 = b64;
  }

  /**
   * Interne Methode, wird von Anwendungen normalerweise nicht aufgerufen.
   * Sie serialisiert das MIME-Boundary  des Attachment.
   *
   *@param  out Outputstream, in den das Attachment serialisiert wird
   *@exception  IOException
   */
  protected void writeXML(java.io.OutputStream out) throws java.io.IOException
  {
    out.write(("\r\n--" + boundary_string + "\r\nContent-Type: " + contentType + "\r\nContent-Transfer-Encoding: ").getBytes(Constants.CHAR_ENCODING));
    out.write((base64 ? "base64" : "8-bit").getBytes(Constants.CHAR_ENCODING));
    out.write(("\r\nContent-ID: <" + getRefID() + ">\r\nContent-Length: " + getLength() + "\r\n").getBytes(Constants.CHAR_ENCODING));

    if (mimeHeaders != null)
    {
      String key;
      Hashtable mh = (Hashtable) mimeHeaders.clone();
      removeHeaders(mh);

      for (Enumeration e = mh.keys(); e.hasMoreElements();)
      {
        key = (String) e.nextElement();
        out.write((key.trim() + ": " + ((String) mh.get(key)).trim() + "\r\n").getBytes(Constants.CHAR_ENCODING));
      }
    }

    out.write(("\r\n").getBytes(Constants.CHAR_ENCODING));

    if (log.isDebugEnabled())
      log.debug("######### SwapBuffero#### " + swapBuffer);

    InputStream in = swapBuffer.getInputStream();
    in.reset();

    if (base64)
      out = new Base64OutputStream(out, true);

    int count = 0;
    byte[] bytesIn = new byte[1024];

    while ((count = in.read(bytesIn)) > -1)
      out.write(bytesIn, 0, count);

    in.close();

    if (base64)
      out.flush();
  }

  /**
   * Liefert true, wenn es sich um ein verschlsseltes Attachment handelt.
   *
   * @return verschlsselt -> true, unverschlsselt -> false
   */
  public boolean isEncrypted()
  {
    return encrypt;
  }

  /**
   * Liefert den Hashwert nach der Verschlsselung fr die Nachrichtensignatur.
   *
   * @return Hashwert
   */
  public byte[] getEncryptedDigestValue(String digestAlgorithm) throws IOException, NoSuchAlgorithmException
  {
	if (encryptedDigestValues.get(digestAlgorithm) == null)
		encryptedDigestValues.put(digestAlgorithm, createDigest(digestAlgorithm, false));

	if (log.isDebugEnabled())
		log.debug(new String((byte[])encryptedDigestValues.get(digestAlgorithm)));
  
	return (byte[])encryptedDigestValues.get(digestAlgorithm);
  }

  /**
   * Liefert eine Hashtable mit den MIME-Headereintrgen des MIME-boundary.
   * Nicht enthalten sind die Headereintrge "Content-Transfer-Encoding",
   * "Content-ID", "Content-Length" und "Content-Type".
   * Abschnitts.
   * @return Header
   */
  public Hashtable getMimeHeaders()
  {
    return mimeHeaders;
  }

  /**
   * Setzt zustzliche MIME-Headereintrge fr den MIME-boundary
   * Abschnitt des Attachments. Die key- und value-Strings der Hashtable mssen
   * den MIME-Spezifikationen gengen. Die Header
   * Content-Transfer-Encoding, Content-ID, Content-Length und Content-Type werden
   * von der Implementierung gesetzt und werden ignoriert.
   * @param headers MIME-Header als key-value-Paare
   * @see #setContentType(String)
   */
  public void setMimeHeaders(Hashtable headers)
  {
    removeHeaders(headers);
    mimeHeaders = headers;
  }

  private void removeHeaders(Hashtable ht)
  {
    String[] hd = new String[] { "Content-Transfer-Encoding", "Content-ID", "Content-Length", "Content-Type" };
    String key;

    for (Enumeration e = ht.keys(); e.hasMoreElements();)
    {
      key = (String) e.nextElement();

      for (int i = 0; i < hd.length; i++)
      {
        if (key.equalsIgnoreCase(hd[i]))
          ht.remove(key);
      }
    }
  }
}
