package de.osci.osci12.messageparts;

import de.osci.helper.Base64InputStream;
import de.osci.helper.Base64OutputStream;
import de.osci.helper.NullOutputStream;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.encryption.Crypto;
import de.osci.osci12.extinterfaces.OSCIDataSource;
import de.osci.osci12.extinterfaces.OSCIDataSourceExt123;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;


/**
 * <p>Die Content-Klasse reprsentiert einen Content-Eintrag in einer OSCI-
 * Nachricht. Die Content-Eintrge befinden sich in ContentContainer-Eintrgen
 * und enthalten die eigentlichen Inhaltsdaten, die in beliebigen Daten,
 * Refenrenzen auf Attachments oder wiederum in Inhaltsdatencontainern bestehen
 * knnen.</p>
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Ezzntwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 */
public class Content extends MessagePart
{
  private static Log log = LogFactory.getLog(Content.class);
  private static int id = -1;
  private Attachment attachment = null;
  protected OSCIDataSource swapBuffer;
  private ContentContainer coco;
  private int contentType;
  private InputStream transformedDataStream;
  public static final int ATTACHMENT_REFERENCE = 0;
  public static final int CONTENT_CONTAINER = 1;
  public static final int DATA = 2;
  String coNS;

  /**
   * Legt ein Content-Objekt an. Die Daten werden aus dem bergebenen
   * InputStream gelesen.
   * @param ins InputStream
   * @throws IOException bei Lesefehlern
   */
  public Content(InputStream ins) throws IOException
  {
    if (ins == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " ins");

    id++;
    this.setRefID("content" + id);
    contentType = DATA;
    load(ins);
    transformers.add(b64);
  }
  
  /**
   * Legt ein Content-Objekt an. Die Daten werden aus dem bergebenen
   * InputStream gelesen. bergebene Transformer-Strings werden in die
   * XML-Signatur eingetragen, die Strings mssen die gesamten Transformer-Eintrge
   * gem der XML-Signature-Spezifikation in kanonischer Form enthalten.
   * Die zu signierenden transformierten Daten werden aus dem als dritten
   * Parameter bergebenen Stream gelesen.
   * @param ins InputStream der Inhaltsdaten
   * @param transformer Array der TransformerEintrge
   * @param transformedData InputStream der transformierten Daten
   * @throws IOException bei Lesefehlern
   * @throws OSCIException bei Problemn beim Aufbau des OSCI-Signatureintrags
   * @throws NoSuchAlgorithmException wenn der verwendete Security-Provider
   * den erforderlichen Hash-Algorithmus nicht untersttzt.
   */
  public Content(InputStream ins, String[] transformer, InputStream transformedData)
          throws IOException, 
                 OSCIException, 
                 NoSuchAlgorithmException
  {
    this(ins);

    if ((transformer != null) && (transformer.length > 0))
    {
      transformedDataStream = transformedData;

      for (int i = 0; i < transformer.length; i++)
        transformers.add(transformer[i]);
    }
  }

  /*
   * Legt ein Content-Objekt an. Die Daten werden aus dem bergebenen
   * OSCIDataSource-Objekt gelesen.
   * @param ins der InputStream
   */
  Content(de.osci.osci12.extinterfaces.OSCIDataSource dataSource)
  {
    if (log.isDebugEnabled())
      log.debug("Konstruktor");

    if (dataSource == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " dataSource");

    id++;
    this.setRefID("content" + id);
    swapBuffer = dataSource;
    contentType = DATA;
  }

  /**
   * Legt ein Content-Objekt mit dem Inhalt des bergebenen Strings an.
   * <b>Hinweis:</b> Die bergabe von Inhaltsdaten als String ist zur 
   * Bequemlichkeit fr Daten vorgesehen, die empfngerseitig ebenfalls
   * als String der OSCI-Nachricht entnommen und weiterverarbeitet werden.
   * Dies ist zu beachten, wenn z.B. XML-Dokumente bergeben werden, die
   * eine Zeichensatzdeklaration beinhalten und mglicherweise empfngerseitig
   * als Binrdaten einem Parser bergeben werden. Die OSCI-Bibliothek
   * bertrgt die Daten UTF-8-codiert.   
   * @param data der Inhalt
   */
  public Content(String data) throws IOException
  {
    if (data == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " data");

    id++;
    this.setRefID("content" + id);
    load(new ByteArrayInputStream(data.getBytes(Constants.CHAR_ENCODING)));
    contentType = DATA;
    transformers.add(b64);
  }

  /**
   * Legt ein Content-Objekt mit dem Inhalt des bergebenen Strings an.
   * bergebene Transformer-Strings werden in die XML-Signatur eingetragen,
   * die Strings mssen die gesamten Transformer-Eintrge gem der
   * XML-Signature-Spezifikation in kanonischer Form enthalten.
   * Die zu signierenden transformierten Daten mssen in dem als dritten
   * Parameter bergebenen String enthalten sein.
   * <b>Man beachte den Hinweis unter {@link #Content(String)} zur
   * bergabe von Inhaltsdaten als String.</b>
   * @param data Inhaltsdaten
   * @param transformer Array der Transformer-Eintrge
   * @param transformedData String mit den transformierten Daten
   * @throws IOException bei Lesefehlern
   * @throws OSCIException bei Problemn beim Aufbau des OSCI-Signatureintrags
   * @throws NoSuchAlgorithmException wenn der verwendete Security-Provider
   * den erforderlichen Hash-Algorithmus nicht untersttzt.
   */
  public Content(String data, String[] transformer, String transformedData)
          throws IOException, 
                 OSCIException, 
                 NoSuchAlgorithmException
  {
    this(data);

    if ((transformer != null) && (transformer.length > 0))
    {
      transformedDataStream = new ByteArrayInputStream(transformedData.getBytes(Constants.CHAR_ENCODING));

      for (int i = 0; i < transformer.length; i++)
        transformers.add(transformer[i]);
    }
  }

  /**
   * Legt ein Content-Objekt an, welches eine Referenz auf ein Attachment
   * enthlt.
   * @param attachment das Attachmentobjekt
   */
  public Content(Attachment attachment)
  {
    transformers.add(can);

    if (attachment == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " attachment");

    id++;
    this.attachment = attachment;
    this.setRefID("content" + id);
    contentType = ATTACHMENT_REFERENCE;
  }

  /**
   * Legt ein Content-Objekt an, welches ein ContentContainer-Objekt enthlt.
   * Eine solche Verschachtelung von ContentContainern ist z.B. fr serielle
   * Signaturen erforderlich.
   * @param contentContainer der Inhaltsdatencontainer
   */
  public Content(ContentContainer contentContainer)
  {
    transformers.add(can);

    if (contentContainer == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " contentContainer");

    id++;
    this.coco = contentContainer;
    this.setRefID("content" + id);
    contentType = CONTENT_CONTAINER;
  }

  private void load(InputStream input) throws IOException
  {
    if (swapBuffer == null)
    {
      swapBuffer = DialogHandler.getNewDataBuffer();

      if (swapBuffer instanceof OSCIDataSourceExt123)
        ((OSCIDataSourceExt123) swapBuffer).setConfidential(true);

      Base64OutputStream b64out = new Base64OutputStream(swapBuffer.getOutputStream(), false);
      byte[] bytes = new byte[1024];
      int anz = 0;

      while ((anz = input.read(bytes)) > -1)
      {
        b64out.write(bytes, 0, anz);
      }

      b64out.close();
      input.close();
    }
  }

  /**
   * Bevor eine Signaturprfung an dem ContentContainer-Objekt durchgefhrt
   * werden kann, welches dieses Content-Objekt enthlt, mssen mit dieser
   * Methode die transformierten Daten bergeben werden. Dies betrifft nur
   * Content-Objekte, die unter Anwendung von Transformationen signiert wurden.
   * @param transformedData transformierte Daten
   * @see #setTransformedData(String transformedData)
   * @see ContentContainer#checkSignature(de.osci.osci12.roles.Role)
   */
  public void setTransformedData(InputStream transformedData)
  {
    if (getTransformerForSignature() == null)
      throw new IllegalStateException(DialogHandler.text.getString("no_transformer_state"));

    if (transformedDataStream != null)
      throw new IllegalStateException(DialogHandler.text.getString("illegal_change_of_transformed_data"));

    transformedDataStream = transformedData;
  }

  /**
   * Bevor eine Signaturprfung an dem ContentContainer-Objekt durchgefhrt
   * werden kann, welches dieses Content-Objekt enthlt, mssen mit dieser
   * Methode die transformierten Daten bergeben werden. Dies betrifft nur
   * Content-Objekte, die unter Anwendung von Transformationen signiert wurden.
   * @param transformedData transformierte Daten
   * @see #setTransformedData(InputStream transformedData)
   * @see ContentContainer#checkSignature(de.osci.osci12.roles.Role)
   */
  public void setTransformedData(String transformedData)
  {
    try
    {
      setTransformedData(new ByteArrayInputStream(transformedData.getBytes(Constants.CHAR_ENCODING)));
    }
    catch (UnsupportedEncodingException ex)
    {
      // Kann bei UTF-8 nicht auftreten
    }
  }

  /**
   * Liefert die TransformerEintrge in der Signatur.
   * @return String-Array mit den TransformerEintrgen
   */
  public String[] getTransformerForSignature()
  {
    if (transformers.size() < 2)
      return null;

    String[] ret = new String[transformers.size() - 1];

    for (int i = 1; i < transformers.size(); i++)
      ret[i - 1] = (String) transformers.get(i);

    return ret;
  }

  /**
   *  Liefert den Hashwert des Eintrags.
   *
   *@return    The digestValue value
   */
  protected byte[] getDigestValue(String digestAlgorithm) throws NoSuchAlgorithmException, 
                                           IOException, 
                                           OSCIException
  {
    if (digestValues.get(digestAlgorithm) != null)
      return (byte[])digestValues.get(digestAlgorithm);

    MessageDigest md;

    if (DialogHandler.getSecurityProvider() == null)
      md = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(digestAlgorithm));
    else
      md = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(digestAlgorithm),
                                     DialogHandler.getSecurityProvider());

    NullOutputStream nos = new NullOutputStream();
    DigestOutputStream digestOut = new DigestOutputStream(nos, md);

    if (contentType == DATA)
    {
      if (transformedDataStream == null)
      {
        // Bibliothek akzeptiert nur Base64-codierte Daten. Workaround fr den Fall,
        // dass dieser Transformer beim Laden zwischengespeicherter Daten verloren gegangen ist (StoredMessage).
        // Signaturen mit Transformern knnen in diesem Fall nur mit Hilfe neuer Content-Instanzen
        // erzeugt werden.
        if (transformers.size() == 0)
          transformers.add(b64);

        if (transformers.size() > 1)
          throw new IllegalStateException(DialogHandler.text.getString("no_transformed_data"));

        swapBuffer.getInputStream().reset();
        transformedDataStream = new Base64InputStream(swapBuffer.getInputStream());
      }

      if (log.isDebugEnabled() && (transformers != null))
      {
        for (int i = 0; i < transformers.size(); i++)
          log.debug("Transfromer " + getRefID() + " - " + transformers.get(i));
      }

      byte[] tmp = new byte[1024];
      int i;

      while ((i = transformedDataStream.read(tmp)) > -1)
      {
        digestOut.write(tmp, 0, i);
      }

      transformedDataStream.close();
      digestOut.close();
      digestValues.put(digestAlgorithm, md.digest());
    }
    else
    {
      writeXML(digestOut, false);
      digestOut.close();
      digestValues.put(digestAlgorithm, md.digest());

      if (log.isDebugEnabled())
        log.debug("enter getDigestValue" + Crypto.toHex((byte[])digestValues.get(digestAlgorithm)));

      length = nos.getLength();
    }

    return (byte[])digestValues.get(digestAlgorithm);

  }

  /**
   * Liefert den InputStream der Daten zurck, wenn der Content einer
   * empfangenen Nachricht entnommen wurde.
   * @return den InputStream der Inhaltsdaten, oder null, wenn die Nachricht
   * einen ContentContainer oder eine Attachmentreferenz enthlt.
   * @throws IOException bei Lesefehlern
   */
  public InputStream getContentStream() throws IOException
  {
    if (swapBuffer != null)
    {
      swapBuffer.getInputStream().reset();

      return new Base64InputStream(swapBuffer.getInputStream());
    }
    else
    {
      return null;
    }
  }

  /**
   * Liefert die eingestellten Daten des Content als String zurck.
   * Der zurckgegebene String ist eine UTF-8-Interpretation der enthaltenen
   * Binrdaten.
   * @return den String der Inhaltsdaten oder null, wenn die Nachricht
   * einen ContentContainer oder eine Attachmentreferenz enthlt
   * @throws IOException bei Lesefehlern
   */
  public String getContentData() throws IOException
  {
    if (swapBuffer != null)
    {
      swapBuffer.getInputStream().reset();

      InputStream in = new Base64InputStream(swapBuffer.getInputStream());
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      byte[] bytes = new byte[1024];
      int anz;

      while ((anz = in.read(bytes)) > -1)
        bos.write(bytes, 0, anz);

      in.close();
      bos.close();

      return bos.toString(Constants.CHAR_ENCODING);
    }
    else

      return null;
  }

  /**
   * Liefert das referenzierte Attachment zurck.
   * @return Attachment-Objekt oder null, wenn das Content-Objekt Nutzdaten
   * oder einen Inhaltsdatencontainern enthlt
   */
  public Attachment getAttachment()
  {
    return attachment;
  }

  /**
   * Liefert den Inhaltsdatencontainer zurck.
   *
   * @return ContentContainer-Objekt oder null, wenn das Content-Objekt
   * Nutzdaten oder eine Referenz auf ein Attachment enthlt
   */
  public ContentContainer getContentContainer()
  {
    return coco;
  }

  /**
   * Gibt die Art des Inhalts des Content-Objektes in Form eines Identifiers
   * zurck. mgliche Werte sind ATTACHMENT_REFERENCE, CONTENT_CONTAINER und DATA
   * @return Inhaltstyp
   */
  public int getContentType()
  {
    return contentType;
  }

  /**
   * undocumented
   *
   * @return undocumented
   */
  public String toString()
  {
    ByteArrayOutputStream out = new ByteArrayOutputStream();

    try
    {
      writeXML(out);
    }
    catch (Exception ex)
    {
      if (log.isDebugEnabled())
        log.debug("Fehler beim writeXML", ex);
    }

    try
    {
      return out.toString(Constants.CHAR_ENCODING);
    }
    catch (UnsupportedEncodingException ex)
    {
      // kann nicht vorkommen
      return "";
    }
  }

  /**
   * undocumented
   *
   * @param out undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIException undocumented
   */
  protected void writeXML(OutputStream out) throws IOException, 
                                                   OSCIException
  {
    writeXML(out, false);
  }

  /**
   *  Interne Methode, wird von Anwendungen normalerweise nicht aufgerufen.
   *
   *@param  out
   *@exception  IOException
   */
  public void writeXML(OutputStream out, boolean inner)
                throws IOException, 
                       OSCIException
  {
    out.write(("<" + osciNSPrefix + ":Content").getBytes(Constants.CHAR_ENCODING));

    if (attachment != null)
    {
      if (log.isDebugEnabled())
        log.debug("SCHREIBE Attachment." + getRefID());

      if (!inner)
      {
        if (coNS == null)
          out.write(ns);
        else
          out.write(coNS.getBytes(Constants.CHAR_ENCODING));
      }

      out.write((" Id=\"" + getRefID() + "\" href=\"cid:" + attachment.getRefID() + "\"></" + osciNSPrefix +
                ":Content>").getBytes(Constants.CHAR_ENCODING));
    }
    else if (coco != null)
    {
      if (log.isDebugEnabled())
        log.debug("Schreibe ContentContainer." + getRefID());

      if (!inner)
        out.write(ns);

      out.write((" Id=\"" + getRefID() + "\">").getBytes(Constants.CHAR_ENCODING));
      coco.writeXML(out, true);
      out.write(("</" + osciNSPrefix + ":Content>").getBytes(Constants.CHAR_ENCODING));
    }
    else
    {
      if (log.isDebugEnabled())
        log.debug("Schreibe Daten.");

      out.write(("><" + osciNSPrefix + ":Base64Content").getBytes(Constants.CHAR_ENCODING));

      if (!inner)
        out.write(ns);

      out.write((" Id=\"" + getRefID() + "\">").getBytes(Constants.CHAR_ENCODING));

      int count;
      byte[] inBytes = new byte[1024];
      InputStream in = swapBuffer.getInputStream();
      in.reset();

      while ((count = in.read(inBytes)) > -1)
        out.write(inBytes, 0, count);

      in.close();
      out.flush();
      out.write(("</" + osciNSPrefix + ":Base64Content></" + osciNSPrefix + ":Content>").getBytes(Constants.CHAR_ENCODING));
    }
  }
}
