package de.osci.osci12.messageparts;

import de.osci.helper.Base64;
import de.osci.helper.Canonizer;
import de.osci.helper.SymCipherInputStream;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.common.OSCICancelledException;
import de.osci.osci12.encryption.Crypto;
import de.osci.osci12.encryption.OSCICipherException;
import de.osci.osci12.messagetypes.OSCIMessage;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Role;
import de.osci.osci12.signature.KeyInfo;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import java.security.NoSuchAlgorithmException;

import java.util.Hashtable;
import java.util.Vector;

import javax.xml.parsers.ParserConfigurationException;


/**
 * <p>Die EncryptedDataOSCI-Klasse stellt einen Datencontainer fr verschlsselte
 * Daten in einer OSCI-Nachricht dar. Ein EncryptedDataOSCI-Objekt wird in eine
 * OSCI-Nachricht eingestellt oder in einen Content-Container(bei Mehrfachverschlsselung).
 * Anwendungen knnen in ein EncryptedDataOSCI-Objekt nur Content-Container
 * (zum Verschlsseln) einstellen.</p>
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 * @see ContentContainer
 * @see  Attachment
 */
public class EncryptedDataOSCI extends MessagePart
{
  private static Log log = LogFactory.getLog(EncryptedDataOSCI.class);
  private static int idNr = -1;
  private int keyIDs = 0;
  /**
   * EncryptedData-Objekt enthlt eine Referenz auf ein Attachment
   */
  private static final int TYPE_OF_DATA_ATTACHMENT = 0;
  /**
   * EncryptedData Objekt enthlt einen ContentContainer
   */
  private static final int TYPE_OF_DATA_CONTENTCONTAINER = 1;
  private int typeOfData = -1;
  private static int ENCRYPTEDDATA_START = 0;
  private static int ENCRYPTEDDATA_ENCRYPTED = 1;
  private int stateOfObject = ENCRYPTEDDATA_START;
  private de.osci.osci12.encryption.EncryptedData encryptedDataObject = null;
  Vector roles = new Vector();
  Vector readers = new Vector();
  Vector attachments = new Vector();
  private Hashtable encryptedKeyList = new Hashtable();
  private MessagePart content = null;
  private OSCIMessage msg = null;
  private javax.crypto.SecretKey secretKey = null;

  /**
   * Interner Konstruktor fr das Hinzufgen von verschlsselten Attachments zur Nachricht
   * @param attachment
   * @throws IllegalArgumentException
   * @throws IOException
   */
  EncryptedDataOSCI(Attachment attachment) throws IllegalArgumentException, 
                                                  OSCICipherException, 
                                                  IOException
  {
    typeOfData = TYPE_OF_DATA_ATTACHMENT;

    if (attachment == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_thirdargument") + " attachment = null");

    typeOfData = TYPE_OF_DATA_ATTACHMENT;
    this.attachments.add(attachment);

    if (!attachment.isEncrypted())
      throw new IllegalArgumentException(DialogHandler.text.getString("error_unencrypted_attachment"));

    this.secretKey = attachment.secretKey;

    if (log.isDebugEnabled())
      log.debug("Secret-Key des Attachments wird verwendet.");

    de.osci.osci12.encryption.CipherReference ref = new de.osci.osci12.encryption.CipherReference("cid:" +
                                                                                                  attachment.getRefID());
    setRefID("Attachment" + attachment.getRefID());
    encryptedDataObject = new de.osci.osci12.encryption.EncryptedData(ref, Crypto.getCipherAlgoID(this.secretKey),
                                                                      getRefID());
    this.stateOfObject = ENCRYPTEDDATA_ENCRYPTED;

    KeyInfo keyInfo = new KeyInfo();
    keyInfo.setMgmtData(Base64.encode(attachment.secretKey.getEncoded()));
    encryptedDataObject.setKeyInfo(keyInfo);
    this.content = attachment;
  }

  /**
   * Legt ein EncryptedData-Objekt fr TripleDES-Verschlsselung an, welches
   * als zu verschlsselnden Inhalt das bergebenen ContentContainer-Objekt
   * enthlt. Es wird fr die symmetrische Verschlsselung ein
   * AES-256-Schlssel erzeugt.
   *
   * @param coco Inhaltsdatencontainer mit den zu verschlsselnden Daten
   * @throws NoSuchAlgorithmException wenn ein nicht untersttzter Algorothmus
   * bergeben wurde
   */
  public EncryptedDataOSCI(ContentContainer coco) throws NoSuchAlgorithmException, 
                                                         IOException
  {
    this(null, Constants.SYMMETRIC_CIPHER_ALGORITHM_AES256, coco);
  }

  /**
   * Legt ein EncryptedData-Objekt mit dem bergebenen symmetrischen Schlssel
   * an, welches als zu verschlsselnden Inhalt das bergebenen
   * ContentContainer-Objekt enthlt.
   *
   * @param secretKey geheimer Schlssel zum Verschlsseln der Daten
   * @param coco Inhaltsdatencontainer mit den zu verschlsselnden Daten
   * @throws NoSuchAlgorithmException wenn ein nicht untersttzter Algorothmus
   * bergeben wurde
   */
  public EncryptedDataOSCI(javax.crypto.SecretKey secretKey, ContentContainer coco)
                    throws NoSuchAlgorithmException, 
                           IOException
  {
    this(secretKey, null, coco);
  }

  /**
   * Legt ein EncryptedData-Objekt fr den angegebenen symmetrischen
   * Verschlsselungsalgorithmus an, welches als zu verschlsselnden Inhalt
   * das bergebene ContentContainer-Objekt enthlt. Es wird fr die symmetrische
   * Verschlsselung ein Schlssel fr den gewhlten Algorithmus erzeugt.
   *
   * @param symmetricCipherAlgorithm Verschlsselungsalgorithmus, erlaubte Werte
   * sind Constants.SYMMETRIC_CIPHER_ALGORITHM_TDES_CBC und
   * Constants.SYMMETRIC_CIPHER_ALGORITHM_AES128, -192, -265
   * @param coco Inhaltsdatencontainer mit den zu verschlsselnden Daten
   * @throws NoSuchAlgorithmException wenn ein nicht untersttzter Algorithmus
   * bergeben wurde
   */
  public EncryptedDataOSCI(String symmetricCipherAlgorithm, ContentContainer coco)
                    throws NoSuchAlgorithmException, 
                           IOException
  {
    this(null, symmetricCipherAlgorithm, coco);
  }

  private EncryptedDataOSCI(javax.crypto.SecretKey secretKey, String algo, ContentContainer coco)
                     throws NoSuchAlgorithmException
  {
    transformers.add(can);
    id = typ + (++idNr);

    if (secretKey == null)
      secretKey = Crypto.createSymKey(algo);

    algo = secretKey.getAlgorithm();

    if (!algo.equals("DESede") && !algo.equals("AES"))
      throw new NoSuchAlgorithmException(DialogHandler.text.getString("encryption_algorithm_not_supported") + algo);

    if (algo.equals("AES"))
    {
      int len = secretKey.getEncoded().length;

      if ((len != 16) && (len != 24) && (len != 32))
        throw new NoSuchAlgorithmException(DialogHandler.text.getString("encryption_algorithm_not_supported" + " AES-" +
                                                                        (len * 8)));
    }

    if (coco == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_thirdargument") + " coco = null");

    typeOfData = TYPE_OF_DATA_CONTENTCONTAINER;
    id = "encdata_" + idNr++;
    this.secretKey = secretKey;
    this.content = coco;
    encryptedDataObject = new de.osci.osci12.encryption.EncryptedData(coco, Crypto.getCipherAlgoID(this.secretKey), id,
                                                                      this.secretKey);

    Attachment[] atts = ((ContentContainer) content).getAttachments();

    if ((atts != null) && (atts.length > 0))
    {
      for (int i = 0; i < atts.length; i++)
        this.attachments.add(atts[i]);
    }

    for (int i = 0; i < coco.roles.size(); i++)
      roles.add(coco.roles.get(i));

    setNSPrefixes(coco.soapNSPrefix, coco.osciNSPrefix, coco.dsNSPrefix, coco.xencNSPrefix, coco.xsiNSPrefix);

    try
    {
      setNS(new String(coco.ns, Constants.CHAR_ENCODING));
    }
    catch (UnsupportedEncodingException ex)
    {
    }
  }

  /**
   * Dieser Konstruktor wird beim Parsen aufgerufen
   * @param encryptedData
   */
  EncryptedDataOSCI(de.osci.osci12.encryption.EncryptedData encryptedData, OSCIMessage osciMsg)
             throws SAXException
  {
    this.encryptedDataObject = encryptedData;
    stateOfObject = ENCRYPTEDDATA_ENCRYPTED;
    this.id = encryptedData.getId();
    msg = osciMsg;

    de.osci.osci12.encryption.EncryptedKey[] encKeys = encryptedData.getKeyInfo().getEncryptedKeys();

    for (int i = 0; i < encKeys.length; i++)
    {
      String uri = (encKeys[i]).getKeyInfo().getRetrievalMethod().getURI();

      //id
      if (uri.startsWith("#"))
        uri = uri.substring(1);

      Role role = osciMsg.getRoleForRefID(uri);

      if (role == null)
        throw new SAXException(DialogHandler.text.getString("no_cipher_cert") + uri);

      roles.add(role);
      readers.add(role);
    }
  }

  void setNS(String ns)
  {
    encryptedDataObject.encNS = ns;
  }

  /**
   * undocumented
   *
   * @param soap undocumented
   * @param osci undocumented
   * @param ds undocumented
   * @param xenc undocumented
   * @param xsi undocumented
   */
  public void setNSPrefixes(String soap, String osci, String ds, String xenc, String xsi)
  {
    encryptedDataObject.soapNSPrefix = soap;
    encryptedDataObject.osciNSPrefix = osci;
    encryptedDataObject.dsNSPrefix = ds;
    encryptedDataObject.xencNSPrefix = xenc;
    encryptedDataObject.xsiNSPrefix = xsi;
  }

  /**
   *<p>Diese Methode entschlsselt den Inhalt des EncryptedDataOSCI-Objektes mit
   * dem bergebenem Role-Objekt.</p>
   * <p>Die Verschlsselung wird aufgehoben, die entschlsselten Informationen geparst und ein
   * ContentContainer-Objekt aufgebaut und zurckgegeben.
   *</p>
   *@param reader enthlt die Entschlsselungsinformationen. Das Role-Objekt muss auf
   *jeden Fall ein Decrypter-Objekt zur Aufhebung der Verschlsselung enthalten.
   *@return den entschlsselten ConentContainer
   */
  public ContentContainer decrypt(Role reader)
                           throws OSCICipherException, 
                                  OSCIRoleException, 
                                  IOException, 
                                  OSCICancelledException, 
                                  SAXException, 
                                  NoSuchAlgorithmException
  {
    if (reader == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " reader = null");

    String rdId = null;

    for (int i = 0; i < roles.size(); i++)
    {
      if (reader.getCipherCertificate().equals(((Role) roles.get(i)).getCipherCertificate()))
      {
        rdId = ((Role) roles.get(i)).getCipherCertificateId();

        break;
      }
    }

    if (rdId == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("no_encryption_for_role"));

    de.osci.osci12.encryption.EncryptedKey encKey = this.encryptedDataObject.findEncrypedKey(rdId);

    if (log.isDebugEnabled())
      log.debug("Anzahl der Empfnger: " + encryptedKeyList.size());

    if ((encKey == null) || !reader.hasCipherPrivateKey())
      throw new IllegalArgumentException(DialogHandler.text.getString("no_encryption_for_role"));

    if (log.isDebugEnabled())
      log.debug("key: " + encKey.getKeyInfo().getRetrievalMethod().getURI());

    // es wurde ein EncryptedKey gefunden
    // Entschlsseln des sym. Schlssels
    InputStream keyIn = encKey.getCipherData().getCipherValue().getCipherValueStream();
    keyIn.reset();

    byte[] dt = de.osci.helper.Tools.readBytes(keyIn);
    byte[] decryptedKey = reader.getDecrypter().decrypt(dt);

    // Unterscheidung ob es sich um eine CipherRef oder Cipher Value handelt
    if (encryptedDataObject.getCipherData().getCipherReference() != null)
    {
      if (log.isDebugEnabled())
        log.debug("Es handelt sich um eine CipherReference unbekannter herkunft.");

      throw new OSCICipherException("invalid_reference");
    }
    else
    {
      if (log.isDebugEnabled())
        log.debug("Es handelt sich um ein CipherValue.");

      InputStream in = encryptedDataObject.getCipherData().getCipherValue().getCipherValueStream();
      in.reset();

      SymCipherInputStream cin = new SymCipherInputStream(in,
                                                          de.osci.osci12.encryption.Crypto.createSymKey(decryptedKey,
                                                                                                        encryptedDataObject.getEncryptionMethodAlgorithm()),
                                                          false);

      return parseInputStream(cin);
    }
  }

  /**
   * Diese Methode erstellt die EncryptedData-Strukturen ohne den symmetrischen Schlssel noch einmal
   * zu verschlsseln. Dies kann bei vielen Verschlsselungen zu einer schnelleren Verarbeitung mit weniger
   * PIN-Eingaben fhren.
   *
   * @param encryptedSymKey verschlsselter symmetrischer Schlssel
   * @param reader Role-Objekt, welches den Leser reprsentiert
   */
  public void encrypt(byte[] encryptedSymKey, Role reader)
               throws OSCIRoleException, 
                      IOException, 
                      OSCICipherException
  {
    if (!readers.contains(reader))
    {
      roles.add(reader);
      this.readers.add(reader);
    }

    this.stateOfObject = ENCRYPTEDDATA_ENCRYPTED;

    if (log.isDebugEnabled())
      log.debug("Encrypted-Data Methode encrypt mit :" + reader.getCipherCertificateId()); //+ " encryptedSymKey: " +

    KeyInfo keyInfo = null;

    if (content == null)
      keyInfo = new KeyInfo("#" + reader.getCipherCertificate());
    else
      keyInfo = new KeyInfo("#" + reader.getCipherCertificateId());

    de.osci.osci12.encryption.CipherValue cipherValue = new de.osci.osci12.encryption.CipherValue(encryptedSymKey);
    de.osci.osci12.encryption.EncryptedKey encryptedKey = new de.osci.osci12.encryption.EncryptedKey(de.osci.osci12.common.Constants.ASYMMETRIC_CIPHER_ALGORITHM_RSA_1_5,
                                                                                                     cipherValue);
    encryptedKey.setId("EncData_" + id + "_" + keyIDs);
    encryptedKey.setKeyInfo(keyInfo);

    if (encryptedDataObject.getKeyInfo() == null)
      encryptedDataObject.setKeyInfo(new KeyInfo());

    encryptedDataObject.getKeyInfo().addEncryptedKey(encryptedKey);
    encryptedKeyList.put(reader, encryptedKey);

    if (content instanceof ContentContainer)
      this.createEncryptedAttachments((ContentContainer) content, encryptedSymKey, reader);

    if (log.isDebugEnabled())
      log.debug("Fertig mit Encrypted-Data Methode encrypt mit.");
  }

  /**
   * Diese Methode parst den soeben entschlsselten Xml-Strom und Liefert das Ergebnisobjekt zurck
   * @param in Stream der geparst wird
   * @return  Erstelltes Object (derzeit nur coco)
   * @throws OSCICipherException
   */
  private ContentContainer parseInputStream(InputStream in)
                                     throws OSCICipherException, 
                                            IOException, 
                                            SAXException, 
                                            NoSuchAlgorithmException
  {
    ContentPackageBuilder copaBuilder = null;

    try
    {
      Canonizer can = new Canonizer(in, null);
      InputStreamReader isr = new InputStreamReader(can, Constants.CHAR_ENCODING);
      javax.xml.parsers.SAXParserFactory saxFactory = javax.xml.parsers.SAXParserFactory.newInstance();
      saxFactory.setNamespaceAware(true);

      javax.xml.parsers.SAXParser parser = saxFactory.newSAXParser();
      XMLReader rd = parser.getXMLReader();
      copaBuilder = new ContentPackageBuilder(rd, msg, can);
      rd.setContentHandler(copaBuilder);
      rd.parse(new InputSource(isr));
    }
    catch (ParserConfigurationException ex1)
    {
      throw new SAXException(ex1);
    }
    finally
    {
      in.close();
    }

    if (copaBuilder.getLastCreatedObject() instanceof ContentContainer)
    {
      ContentContainer coco = (ContentContainer) copaBuilder.getLastCreatedObject();
      decryptAttachments(coco);

      return coco;
    }
    else
      throw new OSCICipherException("sax_exception");
  }

  private void decryptAttachments(ContentContainer cc)
                           throws IOException
  {
    EncryptedDataOSCI[] encDatas = cc.getEncryptedData();

    for (int i = 0; i < encDatas.length; i++)
    {
      if (encDatas[i].encryptedDataObject.getCipherData().getCipherReference() != null) // getRefID().startsWith("Attachment"))
      {
        String uri = encDatas[i].encryptedDataObject.getCipherData().getCipherReference().getURI();

        if (uri.startsWith("cid:"))
          uri = uri.substring(4);

        encDatas[i].setRefID(uri);

        Hashtable msgAtts = msg.attachments;
        Attachment att = (Attachment) msgAtts.get(uri);
        att.stateOfAttachment = Attachment.STATE_OF_ATTACHMENT_ENCRYPTED;
        att.encrypt = true;

        if (log.isDebugEnabled())
          log.debug("das Attachment: " + att + " URI: " + uri);

        att.secretKey = Crypto.createSymKey(Base64.decode(encDatas[i].encryptedDataObject.getKeyInfo().getMgmtData()),
                                            encDatas[i].encryptedDataObject.getEncryptionMethodAlgorithm());
        cc.attachments.put(att.getRefID(), att);
        cc.removeEncryptedData(encDatas[i], false);
      }
    }

    Content[] cnts = cc.getContents();

    for (int i = 0; i < cnts.length; i++)
    {
      if (cnts[i].getContentType() == Content.CONTENT_CONTAINER)
        decryptAttachments(cnts[i].getContentContainer());
    }
  }

  /**
   * Fgt beim Verschlsseln die EncryptedData-Elemente der Attachments hinzu
   * @param encryptedSymKey
   * @param reader
   * @throws OSCICipherException
   * @throws IOException
   */
  private void createEncryptedAttachments(ContentContainer cc, byte[] encryptedSymKey, Role reader)
                                   throws OSCICipherException, 
                                          IOException
  {
    Attachment[] atts = cc.getAttachments();

    if (atts != null)
    {
      ContentContainer coco = (ContentContainer) content;
      coco.stateOfObject = ContentContainer.STATE_OF_OBJECT_PARSING;

      for (int i = 0; i < atts.length; i++)
      {
        if (log.isDebugEnabled())
          log.debug("Ein weiteres Attachment.");

        EncryptedDataOSCI encData = new EncryptedDataOSCI(atts[i]);
        encData.setNSPrefixes(coco.soapNSPrefix, coco.osciNSPrefix, coco.dsNSPrefix, coco.xencNSPrefix, coco.xsiNSPrefix);
        encData.setNS(new String(coco.ns, Constants.CHAR_ENCODING));
        coco.addEncryptedData(encData);
      }

      Role[] roles = ((ContentContainer) content).getRoles();

      for (int i = 0; i < roles.length; i++)
        if (!this.roles.contains(roles[i]))
          this.roles.add(roles[i]);
    }

    Content[] cnts = cc.getContents();

    for (int i = 0; i < cnts.length; i++)
    {
      if (cnts[i].getContentType() == Content.CONTENT_CONTAINER)
        createEncryptedAttachments(cnts[i].getContentContainer(), encryptedSymKey, reader);
    }
  }

  /**
   * Verschlsselt den geheimen Schlssel und fgt ihn als EncryptedKey-Element
   * dem EncyptedData-Element hinzu. Auerdem wird das Rollen-Objekt der Parent-
   * Nachricht (des MessagePart-Objektes) hinzugefgt, so dass das
   * Verschlsselungszertifikat in der Nachricht enthalten ist.
   * @param reader Rollen-Objekt, fr welches verschlsselt werden soll
   * @throws OSCIRoleException wenn dem Rollen-Objekt das erforderliche
   * Verschlsselungszertifikat fehlt
   */
  public void encrypt(de.osci.osci12.roles.Role reader)
               throws OSCICipherException, 
                      OSCIRoleException, 
                      IOException, 
                      NoSuchAlgorithmException
  {
    if (!readers.contains(reader))
    {
      roles.add(reader);
      readers.add(reader);
    }

    this.stateOfObject = ENCRYPTEDDATA_ENCRYPTED;

    if (log.isDebugEnabled())
      log.debug("Encrypted-Data Methode encrypt mit :" + reader.getCipherCertificateId() +
                " Es handelt sich um ein EncryptedData Objekt vom Typ:" + this.typeOfData);

    if (secretKey == null)
      throw new OSCICipherException("no_secret_key");

    byte[] encryptedSymKey = de.osci.osci12.encryption.Crypto.doRSAEncryption(reader.getCipherCertificate(), secretKey);
    encrypt(encryptedSymKey, reader);
  }

  /**
   *  Interne Methode, wird von Anwendungen normalerweise nicht aufgerufen.
   *
   *@param  out
   *@exception  IOException
   *
   */
  protected void writeXML(OutputStream out) throws IOException, 
                                                   OSCIException
  {
    writeXML(out, true);
  }

  /**
   * undocumented
   *
   * @param out undocumented
   * @param inner undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void writeXML(OutputStream out, boolean inner)
                   throws IOException, 
                          OSCIException
  {
    if (stateOfObject < ENCRYPTEDDATA_ENCRYPTED)
      throw new IllegalStateException(DialogHandler.text.getString("invalid_stateofobject") + " not encrypted.");

    encryptedDataObject.setId(getRefID());
    encryptedDataObject.writeXML(out, inner);
  }

  /**
   * Liefert den Identifier des symmetrischen Verschlsselungsalgorithmus.
   *
   * @return Algorithmus
   */
  public String getSymEncryptionMethod()
  {
    return encryptedDataObject.getEncryptionMethodAlgorithm();
  }

  /**
   * Liefert die eingestellten Attachment-Objekte des ContentContainer.
   * @return Array der zugehrigen Attachments
   */
  public Attachment[] getAttachments()
  {
    return (Attachment[]) this.attachments.toArray(new Attachment[0]);
  }

  /**
   * Liefert die eingestellten Role-Objekte des EncryptedData-Objektes,
   * welche fr die Signatur und/oder Verschlsselung verwendet wurden.
   * @return Array der verwendeten Role-Objekte
   */
  public Role[] getRoles()
  {
    return (Role[]) this.roles.toArray(new Role[0]);
  }

  /**
   * Liefert die eingestellten Role-Objekte des EncryptedData-Objektes,
   * mit dem die Daten verschlsselt wurden.
   * @return Array der Verschlsselungs-Role-Objekte
   */
  public Role[] getReaders()
  {
    return (Role[]) readers.toArray(new Role[] {  });
  }
}
