package de.osci.osci12.messageparts;

import de.osci.helper.NullOutputStream;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.encryption.Crypto;
import de.osci.osci12.messagetypes.OSCIMessage;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import java.security.DigestOutputStream;
import java.security.MessageDigest;

import java.util.Hashtable;
import java.util.Vector;


/**
 * <p>Diese Klasse ist die Basisklasse fr smtliche Header-Element und alle weiteren
 * Elemente welche in einer OSCI-Nachricht signiert werden (Attachment, ContentContainer,
 * Content... ).</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 */
public abstract class MessagePart
{
  private static Log log = LogFactory.getLog(MessagePart.class);
  protected java.util.Hashtable xmlns = new java.util.Hashtable();
  protected String id = "";
  protected String typ;
  protected Hashtable digestValues = new Hashtable();
  protected long length = -1;
  protected Vector transformers = new Vector();
  protected static final String b64 = "<ds:Transform Algorithm=\"" + Constants.TRANSFORM_BASE64 + "\"></ds:Transform>";
  protected static final String can = "<ds:Transform Algorithm=\"" + Constants.TRANSFORM_CANONICALIZATION +
                                      "\"></ds:Transform>";
  public byte[] ns;
  protected MessageDigest md;
  protected String soapNSPrefix = "soap";
  protected String osciNSPrefix = "osci";
  protected String dsNSPrefix = "ds";
  protected String xencNSPrefix = "xenc";
  protected String xsiNSPrefix = "xsi";

  /**
   * Creates a new MessagePart object.
   */
  protected MessagePart()
  {
    typ = getClass().getName().substring(getClass().getName().lastIndexOf('.') + 1).toLowerCase();

    try
    {
      ns = " xmlns:ds=\"http://www.w3.org/2000/09/xmldsig#\" xmlns:osci=\"http://www.osci.de/2002/04/osci\" xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xenc=\"http://www.w3.org/2001/04/xmlenc#\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"".getBytes(Constants.CHAR_ENCODING);
    }
    catch (UnsupportedEncodingException ex)
    {
      // Jede JVM untersttzt UTF-8
    }
  }

  /**
   * Setzt die Namespace-Prefixe und die NS-Deklaration, die bei der Serialisierung verwendet werden.
   * Sollte vom Anwender normalerweise nicht aufgerufen werden.
   * @param msg die Nachricht, aus der die Prefixes bernommen werden sollen
   */
  public void setNSPrefixes(OSCIMessage msg)
  {
    soapNSPrefix = msg.soapNSPrefix;
    osciNSPrefix = msg.osciNSPrefix;
    dsNSPrefix = msg.dsNSPrefix;
    xencNSPrefix = msg.xencNSPrefix;
    xsiNSPrefix = msg.xsiNSPrefix;

    try
    {
      ns = msg.ns.getBytes(Constants.CHAR_ENCODING);
    }
    catch (UnsupportedEncodingException ex)
    {
      // Tritt nicht auf
    }
  }

  /**
   * Setzt die Namespace-Prefixe, die bei der Serialisierung verwendet werden.
   * Sollte vom Anwender normalerweise nicht aufgerufen werden.
   * @param soap SOAP-Prefix
   * @param osci OSCI-Prefix
   * @param ds Digital Signature-Prefix
   * @param xenc XML Encryption-Prefix
   * @param xsi XML Schema-Prefix
   */
  public void setNSPrefixes(String soap, String osci, String ds, String xenc, String xsi)
  {
    soapNSPrefix = soap;
    osciNSPrefix = osci;
    dsNSPrefix = ds;
    xencNSPrefix = xenc;
    xsiNSPrefix = xsi;
  }

  /**
   * Setzt das refID-Attribut des MessagePart-Objekts. Diese Eigenschaft wird
   * in Subklassen fr unterschiedliche Zwecke verwendet, z.B. als Referenz-Id
   * fr die Signaturen. Das Attribut muss mindestens innerhalb der OSCI-Nachricht
   * eindeutig sein. 
   *
   * @param  id  refID-String
   * @see #getRefID()
   */
  public void setRefID(String id)
  {
    this.id = id;
  }

  /**
   * Liefert das refID-Attribut.
   * @return refID
   * @see #setRefID(String)
   */
  public String getRefID()
  {
    return id;
  }

  /*
   *  Liefert den Hashwert des Message-Parts.
   *
   *@return    The digestValue value
   */
  protected byte[] getDigestValue(String digestAlgorithm) throws java.security.NoSuchAlgorithmException, 
                                           IOException, 
                                           OSCIException
  {
    if (digestValues.get(digestAlgorithm) != null)
      return (byte[])digestValues.get(digestAlgorithm);

    if (DialogHandler.getSecurityProvider() == null)
      md = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(digestAlgorithm));
    else
      md = MessageDigest.getInstance((String) Constants.JCA_JCE_MAP.get(digestAlgorithm),
                                     DialogHandler.getSecurityProvider());

    NullOutputStream nos = new NullOutputStream();
    DigestOutputStream digestOut = new DigestOutputStream(nos, md);
    writeXML(digestOut);
    digestOut.close();
    digestValues.put(digestAlgorithm, md.digest());

    if (log.isDebugEnabled())
      log.debug("enter getDigestValue" + Crypto.toHex((byte[])digestValues.get(digestAlgorithm)));

    length = nos.getLength();

    return (byte[])digestValues.get(digestAlgorithm);
  }

  /**
   * Berechnet die Lnge des XML-Tags.
   * @return Lnge
   * @throws IOException
   * @throws de.osci.osci12.roles.OSCIRoleException
   */
  public long getLength() throws IOException, 
                                 OSCIException
  {
    if (length > -1)
      return length;

    de.osci.helper.NullOutputStream out = new de.osci.helper.NullOutputStream();
    writeXML(out);
    out.close();
    length = out.getLength();

    return length;
  }

  /**
   * undocumented
   *
   * @param out undocumented
   *
   * @throws IOException undocumented
   * @throws de.osci.osci12.OSCIException undocumented
   */
  protected abstract void writeXML(OutputStream out) throws IOException, 
                                                            de.osci.osci12.OSCIException;
}
