package de.osci.osci12.messageparts;

import de.osci.helper.CanParser;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messagetypes.OSCIMessage;
import de.osci.osci12.messagetypes.OSCIMessageBuilder;
import de.osci.osci12.messagetypes.OSCIRequest;
import de.osci.osci12.roles.Role;
import de.osci.osci12.signature.OSCISignatureException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;


/**
 * Signature-Parser.
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann / N. Bngener
 * @version 1.5
 */
class OSCISignatureBuilder extends MessagePartParser
{
  private static Log log = LogFactory.getLog(OSCISignatureBuilder.class);
  OSCISignature sig;
  private OSCISignatureReference sr;
  private String signingTime;
  private String signingPropsId;

  /**
   *  Constructor for the OSCISignatureBuilder object
   *
   *@param  xmlReader      Description of Parameter
   *@param  parentHandler  Description of Parameter
   *@param  atts           Description of Parameter
   */
  public OSCISignatureBuilder(XMLReader xmlReader, DefaultHandler parentHandler, Attributes atts)
  {
    super(xmlReader, parentHandler);

    if (parentHandler instanceof OSCIMessageBuilder)
    {
      OSCIMessage msg = ((OSCIMessageBuilder) parentHandler).getOSCIMessage();
      String name = "<" + msg.osciNSPrefix + ":";
      String soap = msg.soapNSPrefix;
      name += ((((de.osci.osci12.messagetypes.OSCIMessageBuilder) parentHandler).getOSCIMessage() instanceof OSCIRequest)
               ? "ClientSignature Id=\"" : "SupplierSignature Id=\"");
      name += (atts.getValue("Id") + "\" ");
      name += (soap + ":actor=\"http://schemas.xmlsoap.org/soap/actor/next\" " + soap + ":mustUnderstand=\"1\">");
      sig = new OSCISignature(name);
      sig.setNSPrefixes(soap, msg.osciNSPrefix, msg.dsNSPrefix, msg.xencNSPrefix, msg.xsiNSPrefix);
    }
    else
      sig = new OSCISignature();
  }

  /**
   *  Description of the Method
   *
   *@param  uri                           Description of Parameter
   *@param  localName                     Description of Parameter
   *@param  qName                         Description of Parameter
   *@param  attributes                    Description of Parameter
   *@exception  SAXException  Description of Exception
   */
  public void startElement(String uri, String localName, String qName, Attributes attributes)
                    throws SAXException
  {
    if (log.isDebugEnabled())
      log.debug("Start Element" + uri + ":" + localName);

    if (uri.equals(DS_XMLNS))
    {
      if (localName.equals("Signature")
        || localName.equals("SignedInfo")
        || localName.equals("Object")
        || localName.equals("CanonicalizationMethod"))
        ;
      else if (localName.equals("DigestMethod"))
        sr.setDigestMethodAlgorithm(attributes.getValue("Algorithm"));
      else if (localName.equals("SignatureValue"))
        currentElement = new StringBuffer();
      else if (localName.equals("KeyInfo"))
        currentElement = new StringBuffer();
      else if (localName.equals("SignatureMethod"))
      {
        sig.signatureAlgorithm = attributes.getValue("Algorithm");

        if (Constants.JCA_JCE_MAP.get(attributes.getValue("Algorithm")) == null)
          throw new SAXException(DialogHandler.text.getString("invalid_signature_algorithm") + " " +
                                 attributes.getValue("Algorithm"));
      }
      else if (localName.equals("Reference"))
      {
        sr = new OSCISignatureReference();

        String rf = attributes.getValue("URI");
        sr.setRefID(rf);

        if (log.isDebugEnabled())
          log.debug("############ Anzahl refs Start:" + sig.refs.size());

        sig.refs.add(sr);
      }
      else if (localName.equals("Transforms"))
      {
        // do nothing
      }
      else if (localName.equals("Transform"))
      {
        //      sr.transformerAlgorithms.add(attributes.getValue("Algorithm"));
        try
        {
          CanParser cp = new CanParser(sr.transformerAlgorithms, xmlReader, this, qName);
          xmlReader.setContentHandler(cp);
          cp.startDocument();
          cp.startElement(uri, localName, qName, attributes);
        }
        catch (Exception ex)
        {
          throw new SAXException(DialogHandler.text.getString("sax_exception_transfomation"), ex);
        }
      }
      else if (localName.equals("DigestMethod "))
        sr.setDigestMethodAlgorithm(attributes.getValue("Algorithm"));
      else if (localName.equals("DigestValue") ||
                   localName.equals("SignatureValue"))
        currentElement = new StringBuffer();
      else if (localName.equals("RetrievalMethod"))
      {
        String rf = attributes.getValue("URI");
        sig.signerId = rf;
      }
      else
        throw new SAXException(DialogHandler.text.getString("unexpected_entry") + ": "
                               + uri + ":" + localName);
    }
    else if (uri.equals(XADES_XMLNS))
    {
      if (localName.equals("QualifyingProperties")
        || localName.equals("SignedSignatureProperties"))
        ;
      else if (localName.equals("SigningTime"))
        currentElement = new StringBuffer();
      else if (localName.equals("SignedProperties"))
      {
        signingPropsId = attributes.getValue("Id");
      }
      else
        throw new SAXException(DialogHandler.text.getString("unexpected_entry") + ": " + localName);
    }
    else
      throw new SAXException(DialogHandler.text.getString("unexpected_entry") + ": " + localName);
  }

  /**
   *  Description of the Method
   *
   *@param  uri                           Description of Parameter
   *@param  localName                     Description of Parameter
   *@param  qName                         Description of Parameter
   *@exception  SAXException  Description of Exception
   */
  public void endElement(String uri, String localName, String qName)
                  throws SAXException
  {
    if (log.isDebugEnabled())
      log.debug("End-Element: " + localName);

    try
    {
      if (uri.equals(DS_XMLNS))
      {
        if (localName.equals("Reference"))
        {
          if (log.isDebugEnabled())
          {
            log.debug("############ Anzahl refs End :" + sig.refs.size());

            String[] tr = sr.getTransformerAlgorithms();

            for (int i = 0; i < tr.length; i++)
            {
              log.debug("\nTRANS: " + sr.getRefID() + " - " + tr[i]);
            }
          }

          sr = null;
        }
        else if (localName.equals("DigestValue"))
        {
          if (log.isDebugEnabled())
            log.debug("ID " + sr.getRefID());

          byte[] dv = de.osci.helper.Base64.decode(currentElement.toString());
          sr.setDigestValue(dv);
        }
        else if (localName.equals("SignatureValue"))
          sig.signatureValue = de.osci.helper.Base64.decode(currentElement.toString());
        else if (localName.equals("SignedInfo"))
        {
        }
        else if (localName.equals("Signature"))
        {
          if (parentHandler instanceof ContentContainerBuilder)
          {
            if (log.isDebugEnabled())
              log.debug("##########" + sig.signerId);

            String sigId = sig.signerId;

            if (sigId.charAt(0) == '#')
              sigId = sigId.substring(1);

            Role signer = ((ContentContainerBuilder) parentHandler).msg.getRoleForRefID(sigId);

            if (signer == null)
            {
              log.error("Das referenzierte Signer-Objekt konnte nicht gefunden werden.");
              throw new OSCISignatureException("no_signature_cert", sigId);
            }

            sig.signer = signer;
            ((ContentContainerBuilder) parentHandler).getContentContainer().addSignature(sig);
            sig.signedInfo = (byte[]) ((ContentContainerBuilder) parentHandler).can.getSignedInfos().remove(1);
            if (signingPropsId != null)
            {
                sig.signingProperties = (String)((ContentContainerBuilder) parentHandler).can.getSignedProperties().remove(0);
                sig.signingPropsId = signingPropsId;
                sig.signingTime = signingTime;
            }
          }
          else if (parentHandler instanceof OSCIMessageBuilder)
            ((de.osci.osci12.messagetypes.OSCIMessageBuilder) parentHandler).getOSCIMessage().signatureHeader = sig;

          this.xmlReader.setContentHandler(this.parentHandler);
        }
      }
      else if (localName.equals("SigningTime")
        && uri.equals(XADES_XMLNS)
        && (parentHandler instanceof ContentContainerBuilder))
        {
          signingTime = currentElement.toString();
        }
    }
    catch (Exception ex)
    {
      log.error("Fehler im End-Element!", ex);
      throw new SAXException(ex);
    }

    currentElement = null;
  }
}
