package de.osci.osci12.messageparts;

import de.osci.osci12.common.Constants;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;

import javax.xml.parsers.SAXParser;


/**
 * Diese Klasse bildet den OSCI-Laufzettel ab.
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.5
 */
public class ProcessCardBundle extends MessagePart implements Serializable
{
  private static final long serialVersionUID = -5723258187933747170L;
  private static Log log = LogFactory.getLog(ProcessCardBundle.class);
  public String name;
  private String recentModification;
  private String subject;
  private String messageId;
  Timestamp creation;
  Timestamp forwarding;
  Timestamp reception;
  private Inspection[] inspections;

  ProcessCardBundle(String name)
  {
    this.name = name;
  }

  ProcessCardBundle(String name, String messageId, String recentModification, Timestamp creation, Timestamp forwarding,
                    Timestamp reception, String subject, Inspection[] inspections)
  {
    this.name = name;
    this.messageId = messageId;
    this.recentModification = recentModification;
    this.creation = creation;
    this.forwarding = forwarding;
    this.reception = reception;
    this.subject = subject;
    this.inspections = inspections;
  }

  /**
   * Liefert den Zeitstempel der Erstellung.
   * @return Zeitstempel
   */
  public Timestamp getCreation()
  {
    return creation;
  }

  /**
   * Liefert den Zeitstempel der Weiterleitung.
   * @return Zeitstempel
   */
  public Timestamp getForwarding()
  {
    return forwarding;
  }

  /**
   * Liefert den Zeitstempel des Empfangs.
   * @return Zeitstempel
   */
  public Timestamp getReception()
  {
    return reception;
  }

  /**
   * Liefert die Ergebnisse der Zertifikatsprfungen in Form von Inspection-Objekten,
   * die im ProcessCardBundle-Objekt enthalten sind.
   * @return inspections die Prfergebnisse
   */
  public Inspection[] getInspections()
  {
    return inspections;
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   */
  public String getMessageId()
  {
    return messageId;
  }

  /**
   * Liefert das Datum der letzten nderung des Laufzettels. Das Format
   * entspricht dem XML-Schema nach <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">
   * http://www.w3.org/TR/xmlschema-2/#dateTime</a>.
   * @return Datum der letzten nderung
   */
  public String getRecentModification()
  {
    return recentModification;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   */
  public String getSubject()
  {
    return subject;
  }

  /**
   * Interne Methode.
   * @return ProcessCardBundle als String
   * @throws IOException
   */
  public String writeToString() throws IOException
  {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    writeXML(out);
    out.close();

    return out.toString(Constants.CHAR_ENCODING);
  }

  void setCreation(Timestamp creation)
  {
    this.creation = creation;
  }

  void setForwarding(Timestamp forwarding)
  {
    this.forwarding = forwarding;
  }

  void setInspections(Inspection[] inspections)
  {
    this.inspections = inspections;
  }

  void setMessageId(String messageId)
  {
    this.messageId = messageId;
  }

  void setName(String name)
  {
    this.name = name;
  }

  void setReception(Timestamp reception)
  {
    this.reception = reception;
  }

  void setSubject(String subject)
  {
    this.subject = subject;
  }

  void setRecentModification(String recentModification)
  {
    this.recentModification = recentModification;
  }

  static String encode(String text)
  {
    text = text.replaceAll("&", "&amp;");
    text = text.replaceAll("<", "&lt;");
    text = text.replaceAll(">", "&gt;");

    return text.replaceAll("\r", "&#xD;");
  }

  /**
   * undocumented
   *
   * @param out undocumented
   *
   * @throws IOException undocumented
   */
  protected void writeXML(OutputStream out) throws IOException
  {
    writeXML(out, false);
  }

  /** Interne Methode */
  protected void writeXML(OutputStream out, boolean writeObj)
                   throws IOException
  {
    if (log.isDebugEnabled())
      log.debug("Name: " + name);

    if (log.isDebugEnabled())
      log.debug("MessageID: " + messageId);

    if (log.isDebugEnabled())
      log.debug("recentModi: " + recentModification);

    out.write(("<" + osciNSPrefix + ":" + name).getBytes(Constants.CHAR_ENCODING));

    if (writeObj)
      out.write((" xmlns:" + osciNSPrefix + "=\"http://www.osci.de/2002/04/osci\"").getBytes(Constants.CHAR_ENCODING));

    out.write(("><" + osciNSPrefix + ":MessageId>" +
              de.osci.helper.Base64.encode(messageId.getBytes(Constants.CHAR_ENCODING)) + "</" + osciNSPrefix +
              ":MessageId><" + osciNSPrefix + ":ProcessCard RecentModification=\"" + recentModification + "\">").getBytes(Constants.CHAR_ENCODING));

    if (creation != null)
      creation.writeXML(out);

    if (forwarding != null)
      forwarding.writeXML(out);

    if (reception != null)
      reception.writeXML(out);

    if (subject != null)
      out.write(("<" + osciNSPrefix + ":Subject>" + subject + "</" + osciNSPrefix + ":Subject>").getBytes(Constants.CHAR_ENCODING));

    out.write(("</" + osciNSPrefix + ":ProcessCard><" + osciNSPrefix + ":InspectionReport>").getBytes(Constants.CHAR_ENCODING));

    if (inspections != null)
    {
      for (int i = 0; i < inspections.length; i++)
        inspections[i].writeXML(out);
    }

    out.write(("</" + osciNSPrefix + ":InspectionReport></" + osciNSPrefix + ":" + name + ">").getBytes(Constants.CHAR_ENCODING));
  }

  private void writeObject(java.io.ObjectOutputStream out)
                    throws IOException
  {
    String tmp = this.name;
    name = "ProcessCardBundle";
    this.writeXML(out, true);
    name = tmp;
  }

  private void readObject(java.io.ObjectInputStream in)
                   throws IOException
  {
    XMLReader reader = null;

    try
    {
      javax.xml.parsers.SAXParserFactory saxFactory = javax.xml.parsers.SAXParserFactory.newInstance();
      saxFactory.setNamespaceAware(true);

      SAXParser parser = saxFactory.newSAXParser();
      reader = parser.getXMLReader();
      reader.setFeature("http://xml.org/sax/features/validation", false);

      //      reader.setFeature("http://xml.org/sax/features/namespaces", false);
      int[] check = { 1, -1, -1, -1 };
      this.setName("ProcessCardBundle");

      ProcessCardBundleBuilder processCardBuilder = new ProcessCardBundleBuilder(reader, this, check);
      reader.setContentHandler(processCardBuilder);
      this.setName("ProcessCardBundle");
      reader.parse(new InputSource(in));
    }
    catch (Exception ex)
    {
      throw new IOException(ex.getClass() + ": " + ex.getMessage());
    }
  }
}
