package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Attachment;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.Content;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.messageparts.Inspection;
import de.osci.osci12.messageparts.ProcessCardBundle;
import de.osci.osci12.messageparts.Timestamp;
import de.osci.osci12.roles.Addressee;
import de.osci.osci12.roles.Intermed;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.soapheader.DesiredLanguagesH;
import de.osci.osci12.soapheader.IntermediaryCertificatesH;
import de.osci.osci12.soapheader.OsciH;

import java.io.IOException;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Diese Klasse reprsentiert Nachrichtenobjekte fr Annahmeauftrge.
 * Der Intermedir erzeugt nach dem Erhalt eines Weiterleitungsauftrags eine
 * Instanz dieser Klasse und sendet die Nachricht an den Empfnger (hier als Supplier) .
 * Als Antwort auf diese Nachricht muss der Empfnger ein
 * ResponseToAcceptDelivery-Nachrichtenobjekt mit einer Rckmeldung (Feedback)
 * aufbauen und an den Intermedir zurcksenden.</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 *
 * @see de.osci.osci12.messagetypes.ResponseToAcceptDelivery
 */
public class AcceptDelivery extends OSCIRequest implements ContentPackageI
{
  //  private static Log log = LogFactory.getLog(AcceptDelivery.class);
  public ProcessCardBundle processCardBundle = null;
  public DialogHandler fwdDH;

  /**
   * Konstruktor fr den Supplier Parser
   * @param dh
   */
  AcceptDelivery(DialogHandler dh)
  {
    super(dh);
    messageType = ACCEPT_DELIVERY;
    addressee = ((Addressee) dh.getSupplier());
  }

  AcceptDelivery()
  {
    messageType = ACCEPT_DELIVERY;
  }

  /**
   * Konstruktor fr den Intermedir
   * @param fd
   */
  AcceptDelivery(ForwardDelivery fd) throws OSCIRoleException, 
                                            NoSuchAlgorithmException
  {
    super(new DialogHandler((Intermed) fd.dialogHandler.getSupplier(), fd.addressee,
                            (de.osci.osci12.extinterfaces.TransportI) null));
    messageType = ACCEPT_DELIVERY;
    uriReceiver = fd.uriReceiver;
    addressee = fd.addressee;
    originator = fd.originator;
    desiredLanguagesH = new DesiredLanguagesH(fd.desiredLanguagesH.getLanguageList());

    ContentContainer[] con = fd.getContentContainer();

    for (int i = 0; i < con.length; i++)
      addContentContainer(con[i]);

    EncryptedDataOSCI[] enc = fd.getEncryptedData();

    for (int i = 0; i < enc.length; i++)
      addEncryptedData(enc[i]);

    Attachment[] att = fd.getAttachments();

    for (int i = 0; i < att.length; i++)
      addAttachment(att[i]);

    for (int i = 0; i < fd.getOtherAuthors().length; i++)
      otherAutors.put(fd.getOtherAuthors()[i].id, fd.getOtherAuthors()[i]);

    for (int i = 0; i < fd.getOtherReaders().length; i++)
      otherReaders.put(fd.getOtherReaders()[i].id, fd.getOtherReaders()[i]);

    dialogHandler.setEncryption(fd.getDialogHandler().isEncryption());
    dialogHandler.getControlblock().setResponse(null);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock().setConversationID(null);
    dialogHandler.getControlblock().setSequenceNumber(-1);
    fwdDH = fd.getDialogHandler();
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Diese Methode wirft eine Exception, wenn beim Aufbau oder Versand der
   * Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen
   * dem Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   */
  ResponseToAcceptDelivery send() throws IOException, 
                                         de.osci.osci12.OSCIException, 
                                         NoSuchAlgorithmException
  {
    return (ResponseToAcceptDelivery) transmit(null, null);
  }

  /**
   * Diese Methode liefert den Laufzettel der Zustellung zurck oder null,
   * wenn bei der Verarbeitung der Nachricht ein Fehler aufgetreten ist.
   * Die Informationen im Laufzettel knnen auch direkt ber die einzelnen
   * getX()-Methoden ausgewertet werden.
   *
   * @return den Laufzettel als ProcessCardBundle-Objekt, im Fehlerfall null
   * @see #getTimestampCreation()
   * @see #getTimestampForwarding()
   * @see #getInspections()
   * @see #getSubject()
   * @see #getRecentModification()
   * @see #getMessageId()
   */
  public ProcessCardBundle getProcessCardBundle()
  {
    return processCardBundle;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des Eingangs
   * des Weiterleitungsauftrags beim Intermedir.
   * @return Zeitstempel der Einreichung beim Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampCreation()
  {
    return processCardBundle.getCreation();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des
   * vollstndigen Aufbaus des Annahmeauftrags vom Intermedir fr den Empfnger.
   * @return Zeitstempel der Erstellung des Annahmeauftrags durch den Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampForwarding()
  {
    return processCardBundle.getForwarding();
  }

  /**
   * Liefert die Ergebnisse der Zertifikatsprfungen in Form von Inspection-Objekten,
   * die im ProcessCardBundle-Objekt enthalten sind.
   * @return inspections Prfergebnisse
   * @see #getProcessCardBundle()
   */
  public Inspection[] getInspections()
  {
    return processCardBundle.getInspections();
  }

  /**
   * Liefert den Betreff der Nachricht.
   * @return den Betreff der Zustellung
   */
  public String getSubject()
  {
    return processCardBundle.getSubject();
  }

  /**
   * Liefert das Datum der letzten nderung des Laufzettels. Das Format
   * entspricht dem XML-Schema nach http://www.w3.org/TR/xmlschema-2/#dateTime
   * @return Datum der letzten nderung
   * @see #getProcessCardBundle()
   */
  public String getRecentModification()
  {
    return processCardBundle.getRecentModification();
  }

  /**
   * Liefert die in die Nachricht eingestellten (unverschlsselten) Inhaltsdaten als ContentContainer-Objekte.
   * @return enthaltene ContentContainer mit Inhaltsdaten
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> Inhaltsdaten nach dem ContentContainer
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen ContentContainer oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public ContentContainer getContentContainerByRefID(String refID)
  {
	  return super.getContentContainerByRefID(refID);
  }
  
  /**
   * Durchsucht <b>die unverschlsselten</b> ContentContainer nach dem Content
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen Content oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public Content getContentByRefID(String refID)
  {
	return super.getContentByRefID(refID);  
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten
   * als EncryptedData-Objekte.
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   */
  public String getMessageId()
  {
    return processCardBundle.getMessageId();
  }

  void setProcessCardBundle(ProcessCardBundle processCardBundle)
  {
    this.processCardBundle = processCardBundle;
  }

  void sign()
     throws IOException, 
            OSCIException, 
            de.osci.osci12.common.OSCICancelledException, 
            java.security.NoSuchAlgorithmException
  {
    super.sign();
    // Der IntemedCertHeader liegt hier weiter unten
    messageParts.set(5, intermediaryCertificatesH);
    messageParts.set(3, null);
  }

  /**
   * undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIRoleException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws IOException, 
                                  OSCIException, 
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(3, null);

    if (processCardBundle == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": ProcessCardBundle");

    osciH = new OsciH("acceptDelivery", processCardBundle.writeToString());
    messageParts.add(osciH);

    if (dialogHandler.getClient().hasCipherCertificate())
    {
      if (intermediaryCertificatesH == null)
        intermediaryCertificatesH = new IntermediaryCertificatesH();

      intermediaryCertificatesH.setCipherCertificateIntermediary((Intermed) dialogHandler.getClient());
    }

    messageParts.add(intermediaryCertificatesH);
    createNonIntermediaryCertificatesH();
    messageParts.add(nonIntermediaryCertificatesH);
    messageParts.addAll(customHeaders);

    if (body == null)
      body = new Body(getContentContainer(), getEncryptedData());

    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
