package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.IOException;
import java.io.OutputStream;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte fr Zustellungsabholauftrge
 * angelegt. Clients knnen hiermit maximal eine Nachricht vom Intermedir
 * abrufen.  Als Antwort auf diese Nachricht erhalten sie vom Intermedir
 * ein ResponseToFetchDelivery-Nachrichtenobjekt, welches eine Rckmeldung ber den
 * Erfolg der Operation (getFeedback()) und ggf. die gewnschte Nachricht enthlt.</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 *
 * @see de.osci.osci12.messagetypes.ResponseToFetchDelivery
 */
public class FetchDelivery extends OSCIRequest
{
  private static Log log = LogFactory.getLog(FetchDelivery.class);
  private String selectionRule;
  private int selectionMode = -1;

  FetchDelivery()
  {
    messageType = FETCH_DELIVERY;
  }

  /**
   * Legt ein Nachrichtenobjekt fr einen Zustellungsabholauftrag an.
   *
   * @param dh DialogHandler-Objekt des Dialogs, innerhalb dessen die Nachricht
   * versendet werden soll
   *
   * @see de.osci.osci12.common.DialogHandler
   */
  public FetchDelivery(DialogHandler dh) throws NoSuchAlgorithmException
  {
    super(dh);
    messageType = FETCH_DELIVERY;
    originator = ((Originator) dh.getClient());
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock().setSequenceNumber(dialogHandler.getControlblock().getSequenceNumber() + 1);
  }

  /**
   * Setzt die Auswahlregel fr die abzuholende Nachricht. Der Inhalt des
   * bergebenen Strings hngt vom gewhlten Auswahlmodus ab und kann
   * entweder in einer Base64-codierten Message-Id oder einem Datum bestehen.
   * Das Format eines Datums muss dem XML-Schema nach
   * <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">
   * http://www.w3.org/TR/xmlschema-2/#dateTime</a> (ISO 8601-Format) entsprechen.
   *
   * @param selectionRule Message-Id oder Datum
   * @see #setSelectionMode(int)
   * @see #getSelectionRule()
   * @see #formatISO8601(java.util.Date)
   */
  public void setSelectionRule(String selectionRule)
  {
    this.selectionRule = selectionRule;
  }

  /**
   * Liefert die gesetzte Auswahlregel. Wurde keine Regel gesetzt,
   * wird als default null zurckgegeben.
   * @return Auswahlregel (Message-Id oder Datum)
   * @see #setSelectionRule(String)
   * @see #setSelectionMode(int)
   */
  public String getSelectionRule()
  {
    return selectionRule;
  }

  /**
   * Setzt den Auswahlmodus. Mgliche Werte sind
   * SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION oder NO_SELECTION_RULE
   * (default).
   *
   * @param selectionMode Auswahlmodus
   * @see #SELECT_BY_MESSAGE_ID
   * @see #SELECT_BY_DATE_OF_RECEPTION
   * @see #NO_SELECTION_RULE
   * @see #setSelectionRule(String)
   */
  public void setSelectionMode(int selectionMode)
  {
    if ((selectionMode != SELECT_BY_MESSAGE_ID) && (selectionMode != SELECT_BY_DATE_OF_RECEPTION) &&
            (selectionMode != NO_SELECTION_RULE))
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + selectionMode);

    this.selectionMode = selectionMode;
  }

  /**
   * Liefert den gesetzten Auswahlmodus.
   * @return Auswahlmodus SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION
   * oder NO_SELECTION_RULE
   * @see #setSelectionMode(int)
   * @see #setSelectionRule(String)
   */
  public int getSelectionMode()
  {
    return selectionMode;
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Diese Methode wirft eine Exception, wenn beim Aufbau oder Versand der
   * Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen
   * dem Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send(OutputStream, OutputStream)
   */
  public ResponseToFetchDelivery send() throws IOException, 
                                               de.osci.osci12.OSCIException, 
                                               NoSuchAlgorithmException
  {
    return (ResponseToFetchDelivery) transmit(null, null);
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Die aus- und eingehenden Daten werden zustzlich in die bergebenen
   * Streams geschrieben (unverschlsselte Auftragsdaten). Diese Parameter
   * drfen null sein.
   *
   * @param storeOutput Stream, in den die versendete Nachricht geschrieben wird
   * @param storeInput Stream, in den die empfangene Antwortnachricht geschrieben wird
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send()
   */
  public ResponseToFetchDelivery send(OutputStream storeOutput, OutputStream storeInput)
                               throws IOException, 
                                      de.osci.osci12.OSCIException, 
                                      NoSuchAlgorithmException
  {
    return (ResponseToFetchDelivery) transmit(storeOutput, storeInput);
  }

  void sign()
     throws IOException, 
            OSCIException, 
            de.osci.osci12.common.OSCICancelledException, 
            java.security.NoSuchAlgorithmException
  {
    super.sign();
    // Der NonIntemedCertHeader liegt hier weiter unten
    messageParts.set(5, nonIntermediaryCertificatesH);
    messageParts.set(3, null);
  }

  /**
   * undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIException undocumented
   * @throws NoSuchAlgorithmException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws IOException, 
                                  OSCIException, 
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(3, null);

    if (dialogHandler.getControlblock().getChallenge() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Challenge");

    if (dialogHandler.getControlblock().getResponse() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Response");

    if (dialogHandler.getControlblock().getConversationID() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Conversation-Id");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": SequenceNumber");

    //todo: Evtl. Plausis Selection-Mode / Selection-Rule-Format checken (Datum/MsgID)
    StringBuffer selection = new StringBuffer("<");
    selection.append(osciNSPrefix);
    selection.append(":SelectionRule><");
    selection.append(osciNSPrefix);

    if (selectionMode == SELECT_BY_MESSAGE_ID)
    {
      if (selectionRule == null)
        throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": SelectionRule");

      selection.append(":MessageId>");
      selection.append(de.osci.helper.Base64.encode(selectionRule.getBytes(Constants.CHAR_ENCODING)));
      selection.append("</");
      selection.append(osciNSPrefix);
      selection.append(":MessageId></");
    }
    else if (selectionMode == SELECT_BY_DATE_OF_RECEPTION)
    {
      if (selectionRule == null)
        throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": SelectionRule");

      selection.append(":ReceptionOfDelivery>");
      selection.append(selectionRule);
      selection.append("</");
      selection.append(osciNSPrefix);
      selection.append(":ReceptionOfDelivery></");
    }
    else
      selection.delete(0, selection.length());

    if (selection.length() > 0)
    {
      selection.append(osciNSPrefix);
      selection.append(":SelectionRule>");
    }

    osciH = new OsciH("fetchDelivery", selection.toString());
    messageParts.add(osciH);
    messageParts.add(nonIntermediaryCertificatesH);
    messageParts.addAll(customHeaders);
    body = new Body("");
    //    body.setNSPrefixes(this);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
