package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;

import java.io.IOException;
import java.io.OutputStream;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte fr Laufzettelabholauftrge
 * angelegt. Clients knnen hiermit Laufzettel eingegangener Nachrichten vom Intermedir
 * abrufen.  Als Antwort auf diese Nachricht erhalten sie vom Intermedir
 * ein ResponseToFetchProcessCard-Nachrichtenobjekt, welches eine Rckmeldung ber den
 * Erfolg der Operation und ggf. die gewnschten Laufzettel enthlt.</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 *
 * @see de.osci.osci12.messagetypes.ResponseToFetchProcessCard
 */
public class FetchProcessCard extends OSCIRequest
{
  //  private static Log log = LogFactory.getLog(FetchProcessCard.class);
  private String selectionRule;
  private int selectionMode = -1;
  private int roleForSelection = -1;
  private boolean selectNoReceptionOnly = false;
  private long quantityLimit = -1;

  /**
   * Legt ein Nachrichtenobjekt fr einen Laufzettelabholauftrag an.
   *
   * @param dh DialogHandler-Objekt des Dialogs, innerhalb dessen die Nachricht
   * versendet werden soll
   *
   * @see de.osci.osci12.common.DialogHandler
   */
  public FetchProcessCard(DialogHandler dh) throws NoSuchAlgorithmException
  {
    super(dh);
    originator = ((Originator) dh.getClient());
    messageType = FETCH_PROCESS_CARD;
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock().setSequenceNumber(dialogHandler.getControlblock().getSequenceNumber() + 1);
  }

  FetchProcessCard()
  {
    super();
    messageType = FETCH_PROCESS_CARD;
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Diese Methode wirft eine Exception, wenn beim Aufbau oder Versand der
   * Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen
   * dem Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send(OutputStream, OutputStream)
   */
  public ResponseToFetchProcessCard send() throws IOException, 
                                                  de.osci.osci12.OSCIException, 
                                                  NoSuchAlgorithmException
  {
    return (ResponseToFetchProcessCard) transmit(null, null);
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Die aus- und eingehenden Daten werden zustzlich in die bergebenen
   * Streams geschrieben (unverschlsselte Auftragsdaten). Diese Parameter
   * drfen null sein.
   *
   * @param storeOutput Stream, in den die versendete Nachricht geschrieben wird
   * @param storeInput Stream, in den die empfangene Antwortnachricht geschrieben wird
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send()
   */
  public ResponseToFetchProcessCard send(OutputStream storeOutput, OutputStream storeInput)
                                  throws IOException, 
                                         de.osci.osci12.OSCIException, 
                                         NoSuchAlgorithmException
  {
    return (ResponseToFetchProcessCard) transmit(storeOutput, storeInput);
  }

  /**
   * Setzt die Auswahlregel fr die abzuholende Nachricht. Der Inhalt des
   * bergebenen Strings hngt vom gewhlten Auswahlmodus ab und kann
   * entweder in einer Message-Id oder einem Datum bestehen.
   * Mehrere Message-Ids knnen durch ";" getrennt aneinandergereiht werden.
   * Das Format eines Datums muss dem XML-Schema nach
   * <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">
   * http://www.w3.org/TR/xmlschema-2/#dateTime</a> (ISO 8601-Format) entsprechen.
   *
   * @param selectionRule eine oder mehrere Message-Ids oder Datum
   * @see #setSelectionMode(int)
   * @see #getSelectionRule()
   * @see #formatISO8601(java.util.Date)
   */
  public void setSelectionRule(String selectionRule)
  {
    this.selectionRule = selectionRule;
  }

  /**
   * Liefert die gesetzte Auswahlregel. Der Inhalt des zurckgegebenen
   * Strings hngt vom gesetzten Auswahlmodus ab und kann
   * entweder in einer oder mehrerer Message-Id oder einem Datum bestehen.
   * Mehrere Message-Ids werden aneinandergereiht durch ";" getrennt zurckgegeben.
   * Wurde keine Regel gesetzt, wird als default null zurckgegeben.
   * @return Auswahlregel (Message-Ids oder Datum)
   * @see #setSelectionRule(String)
   * @see #setSelectionMode(int)
   */
  public String getSelectionRule()
  {
    return selectionRule;
  }

  /**
   * Setzt den Auswahlmodus. Mgliche Werte sind
   * SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION,
   * SELECT_BY_RECENT_MODIFICATION oder NO_SELECTION_RULE (default).
   *
   * @param selectionMode Auswahlmodus
   * @see #SELECT_BY_MESSAGE_ID
   * @see #SELECT_BY_DATE_OF_RECEPTION
   * @see #SELECT_BY_RECENT_MODIFICATION
   * @see #NO_SELECTION_RULE
   * @see #setSelectionRule(String)
   */
  public void setSelectionMode(int selectionMode)
  {
    if ((selectionMode != SELECT_BY_MESSAGE_ID) && (selectionMode != SELECT_BY_DATE_OF_RECEPTION) &&
            (selectionMode != SELECT_BY_RECENT_MODIFICATION) && (selectionMode != NO_SELECTION_RULE))
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + selectionMode);

    this.selectionMode = selectionMode;
  }

  /**
   * Legt fest, ob nur Laufzettel von Nachrichten zurckgegeben werden sollen,
   * die an den Absender des Laufzettelabholauftrags geschickt wurden oder
   * solche von Nachrichten, die von dem Absender geschickt wurden. Mgliche Werte sind
   * <ul><li>SELECT_ORIGINATOR - fr Nachrichten, die vom Absender dieses
   * Auftrags gesendet wurden.</li><li>SELECT_ADDRESSEE - fr Nachrichten, die
   * an den Absender dieses Auftrags gesendet wurden.</li><li>SELECT_ALL -
   * fr alle Nachrichten (default).</li></ul> <b>Diese Einstellung ist nur in
   * den Selection-Modes SELECT_BY_DATE_OF_RECEPTION und
   * SELECT_BY_RECENT_MODIFICATION wirksam.</b>
   *
   * @param role Modus
   * @see #SELECT_ORIGINATOR
   * @see #SELECT_ADDRESSEE
   * @see #SELECT_ALL
   * @see #setSelectionRule(String)
   * @see #setSelectionMode(int)
   */
  public void setRoleForSelection(int role)
  {
    if ((role != SELECT_ORIGINATOR) && (role != SELECT_ADDRESSEE) && (role != SELECT_ALL))
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + role);

    this.roleForSelection = role;
  }

  /**
   * Liefert den Identifier fr das Auswahlkriterium, ob nur Laufzettel von
   * Nachrichten zurckgegeben werden sollen, die an den oder vom Absender des
   * Laufzettelabholauftrags geschickt wurden.
   *
   * @see #setRoleForSelection(int)
   * @return Modus
   */
  public int getRoleForSelection()
  {
    return roleForSelection;
  }

  /**
   * Legt fest, ob nur Laufzettel von Nachrichten zurckgegeben werden sollen,
   * fr die keine Empfangsbesttigung vom Empfnger vorliegt.
   *
   * @param noReceptionOnly true -> es werden nur Laufzettel fr nicht zugestellte
   * Nachrichten zurckgegeben. false -> es werden alle Laufzettel zurckgegeben
   * (default). <b>Diese Einstellung ist nur in den Selection-Modes
   * SELECT_BY_DATE_OF_RECEPTION und SELECT_BY_RECENT_MODIFICATION wirksam.</b>
   *
   * @see #setSelectionRule(String)
   * @see #setSelectionMode(int)
   * @see #setRoleForSelection(int)
   */
  public void setSelectNoReceptionOnly(boolean noReceptionOnly)
  {
    this.selectNoReceptionOnly = noReceptionOnly;
  }

  /**
   * Gibt an, ob Liefert den Identifier fr das Auswahlkriterium, ob nur Laufzettel von
   * Nachrichten zurckgegeben werden sollen, die an den oder vom Absender des
   * Laufzettelabholauftrags geschickt wurden.
   *
   * @see #setRoleForSelection(int)
   * @return Modus
   */
  public boolean isSelectNoReception()
  {
    return selectNoReceptionOnly;
  }

  /**
   * Liefert den gesetzten Auswahlmodus.
   * @return den Auswahlmodus SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION,
   * SELECT_BY_RECENT_MODIFICATION oder NO_SELECTION_RULE
   * @see #setSelectionRule(String)
   * @see #setSelectionMode(int)
   */
  public int getSelectionMode()
  {
    return selectionMode;
  }

  /**
   * Liefert die maximale Anzahl zurckzugebender Laufzettel.
   * @return gesetzte maximale Anzahl
   */
  public long getQuantityLimit()
  {
    return quantityLimit;
  }

  /**
   * Legt die maximale Anzahl zurckzugebender Laufzettel fest.
   * @param quantityLimit gewnschte maximale Anzahl
   */
  public void setQuantityLimit(long quantityLimit)
  {
    if (quantityLimit <= 0)
      throw new IllegalArgumentException("Limit fr die Zahl der Laufzettel muss grer 0 sein.");

    this.quantityLimit = quantityLimit;
  }

  /**
   * undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIRoleException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws IOException, 
                                  OSCIException, 
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.addAll(customHeaders);

    if (dialogHandler.getControlblock().getChallenge() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Challenge");

    if (dialogHandler.getControlblock().getResponse() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Response");

    if (dialogHandler.getControlblock().getConversationID() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Conversation-Id");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": SequenceNumber");

    String selectionAttributes = "";

    if (selectNoReceptionOnly)
      selectionAttributes += " NoReception=\"true\"";

    if (roleForSelection == SELECT_ADDRESSEE)
      selectionAttributes += " Role=\"Addressee\"";
    else if (roleForSelection == SELECT_ORIGINATOR)
      selectionAttributes += " Role=\"Originator\"";

    StringBuffer selection = new StringBuffer("<");
    selection.append(osciNSPrefix);
    selection.append(":SelectionRule>");

    if (selectionMode == SELECT_BY_MESSAGE_ID)
    {
      String[] msgIds = selectionRule.split(";");

      for (int i = 0; i < msgIds.length; i++)
      {
        selection.append("<");
        selection.append(osciNSPrefix);
        selection.append(":MessageId>");
        selection.append(de.osci.helper.Base64.encode(msgIds[i].getBytes(Constants.CHAR_ENCODING)));
        selection.append("</");
        selection.append(osciNSPrefix);
        selection.append(":MessageId>");
      }

      selection.append("</");
    }
    else if (selectionMode == SELECT_BY_DATE_OF_RECEPTION)
    {
      selection.append("<");
      selection.append(osciNSPrefix);
      selection.append(":ReceptionOfDelivery" + selectionAttributes + ">" + selectionRule + "</");
      selection.append(osciNSPrefix);
      selection.append(":ReceptionOfDelivery></");
    }
    else if (selectionMode == SELECT_BY_RECENT_MODIFICATION)
    {
      selection.append("<");
      selection.append(osciNSPrefix);
      selection.append(":RecentModification" + selectionAttributes + ">" + selectionRule + "</");
      selection.append(osciNSPrefix);
      selection.append(":RecentModification></");
    }
    else
      selection.delete(0, selection.length());

    if (selection.length() > 0)
    {
      selection.append(osciNSPrefix);
      selection.append(":SelectionRule>");
    }

    if (quantityLimit >= 0)
    {
      selection.append("<");
      selection.append(osciNSPrefix);
      selection.append(":Quantity Limit=\"" + quantityLimit + "\"></");
      selection.append(osciNSPrefix);
      selection.append(":Quantity>");
    }

    body = new Body("<" + osciNSPrefix + ":fetchProcessCard>" + selection + "</" + osciNSPrefix + ":fetchProcessCard>");
    body.setNSPrefixes(this);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
