package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.NonIntermediaryCertificatesH;

import java.io.IOException;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte zur Anforderung einer
 * MessageId angelegt. Clients erhalten als Antwort auf diese Nachricht
 * vom Intermedir ein Nachrichtenobjekt, welches eine Rckmeldung ber den
 * Erfolg der Operation und ggf. die angeforderte MessageId enthlt.</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 *
 * @see de.osci.osci12.messagetypes.ResponseToGetMessageId
 */
public class GetMessageId extends OSCIRequest
{
  //  private static Log log = LogFactory.getLog(GetMessageId.class);
  /**
   * Legt ein Nachrichtenobjekt zur Anforderung einer MessageId an.
   *
   * @param dh DialogHandler-Objekt des Dialogs, innerhalb dessen die Nachricht
   * versendet werden soll
   *
   * @see de.osci.osci12.common.DialogHandler
   */
  public GetMessageId(DialogHandler dh) throws NoSuchAlgorithmException
  {
    super(dh);
    messageType = GET_MESSAGE_ID;
    originator = ((Originator) dh.getClient());

    if (!dialogHandler.isExplicitDialog())
      dialogHandler.resetControlBlock();

    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock().setSequenceNumber(dialogHandler.getControlblock().getSequenceNumber() + 1);
  }

  GetMessageId()
  {
    super();
    messageType = GET_MESSAGE_ID;
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Diese Methode wirft eine Exception, wenn beim Aufbau oder Versand der
   * Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen
   * dem Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   */
  public ResponseToGetMessageId send() throws IOException, 
                                              de.osci.osci12.OSCIException, 
                                              NoSuchAlgorithmException
  {
    return (ResponseToGetMessageId) transmit(null, null);
  }

  /**
   * undocumented
   *
   * @throws OSCIRoleException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws OSCIException, 
                                  NoSuchAlgorithmException, 
                                  IOException
  {
    if (dialogHandler.getClient().hasCipherCertificate())
    {
      if (nonIntermediaryCertificatesH == null)
        nonIntermediaryCertificatesH = new NonIntermediaryCertificatesH();

      nonIntermediaryCertificatesH.setCipherCertificateOriginator((Originator) dialogHandler.getClient());
    }

    super.compose();
    messageParts.addAll(customHeaders);

    if (dialogHandler.getControlblock().getChallenge() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Challenge");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": SequenceNumber");

    body = new Body("<" + osciNSPrefix + ":getMessageId></" + osciNSPrefix + ":getMessageId>");
    body.setNSPrefixes(this);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
