package de.osci.osci12.messagetypes;

import de.osci.helper.Canonizer;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.roles.OSCIRoleException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import java.security.NoSuchAlgorithmException;


/**
 * Nachrichtenparser.
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann / N. Bngener
 * @version 1.5
 */
class OSCIEnvelopeBuilder extends DefaultHandler
{
  final static String XSD_ENC_SIG = "http://www.w3.org/2000/09/xmldsig# oscisig.xsd http://www.w3.org/2001/04/xmlenc# oscienc.xsd";

  // Deklaration der XSD fr die Nachrichtentypen
  final static String XSD_INIT_DIALOG = "http://schemas.xmlsoap.org/soap/envelope/ soapInitDialog.xsd " + XSD_ENC_SIG;
  final static String XSD_RSP_INIT_DIALOG = "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToInitDialog.xsd " +
                                            XSD_ENC_SIG;
  final static String XSD_EXIT_DIALOG = "http://schemas.xmlsoap.org/soap/envelope/ soapExitDialog.xsd " + XSD_ENC_SIG;
  final static String XSD_RSP_EXIT_DIALOG = "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToExitDialog.xsd " +
                                            XSD_ENC_SIG;
  final static String XSD_GET_MSG_ID = "http://schemas.xmlsoap.org/soap/envelope/ soapGetMessageId.xsd " + XSD_ENC_SIG;
  final static String XSD_RSP_GET_MSG_ID = "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToGetMessageId.xsd " +
                                           XSD_ENC_SIG;
  final static String XSD_STORE_DELIVERY = "http://schemas.xmlsoap.org/soap/envelope/ soapStoreDelivery.xsd " +
                                           XSD_ENC_SIG;
  final static String XSD_RSP_STORE_DELIVERY =
    "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToStoreDelivery.xsd " + XSD_ENC_SIG;
  final static String XSD_FETCH_DELIVERY = "http://schemas.xmlsoap.org/soap/envelope/ soapFetchDelivery.xsd " +
                                           XSD_ENC_SIG;
  final static String XSD_RSP_FETCH_DELIVERY =
    "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToFetchDelivery.xsd " + XSD_ENC_SIG;
  final static String XSD_FETCH_PROCESS_CARD = "http://schemas.xmlsoap.org/soap/envelope/ soapFetchProcessCard.xsd " +
                                               XSD_ENC_SIG;
  final static String XSD_RSP_FETCH_PROCESS_CARD =
        "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToFetchProcessCard.xsd " + XSD_ENC_SIG;
  final static String XSD_FORWARD_DELIVERY = "http://schemas.xmlsoap.org/soap/envelope/ soapForwardDelivery.xsd " +
                                             XSD_ENC_SIG;
  final static String XSD_RSP_FORWARD_DELIVERY =
        "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToForwardDelivery.xsd " + XSD_ENC_SIG;
  final static String XSD_ACCEPT_DELIVERY = "http://schemas.xmlsoap.org/soap/envelope/ soapAcceptDelivery.xsd " +
                                            XSD_ENC_SIG;
  final static String XSD_RSP_ACCEPT_DELIVERY =
        "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToAcceptDelivery.xsd " + XSD_ENC_SIG;
  final static String XSD_MEDIATE_DELIVERY = "http://schemas.xmlsoap.org/soap/envelope/ soapMediateDelivery.xsd " +
                                             XSD_ENC_SIG;
  final static String XSD_RSP_MEDIATE_DELIVERY =
        "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToMediateDelivery.xsd " + XSD_ENC_SIG;
  final static String XSD_PROCESS_DELIVERY = "http://schemas.xmlsoap.org/soap/envelope/ soapProcessDelivery.xsd " +
                                             XSD_ENC_SIG;
  final static String XSD_RSP_PROCESS_DELIVERY =
        "http://schemas.xmlsoap.org/soap/envelope/ soapResponseToProcessDelivery.xsd " + XSD_ENC_SIG;
  final static String XSD_MESSAGE_FAULT = "http://schemas.xmlsoap.org/soap/envelope/ soapMessageFault.xsd " +
                                          XSD_ENC_SIG;
  final static String XSD_ENCRYPTED_DATA = "http://schemas.xmlsoap.org/soap/envelope/ soapMessageEncrypted.xsd " +
                                           XSD_ENC_SIG;

  //Allgemeines
  private static Log log = LogFactory.getLog(OSCIEnvelopeBuilder.class);
  protected OSCIMessage msg = null;
  XMLReader xmlReader = null;
  /** Fr internen Gebrauch */
  Canonizer hashNCanStream = null;
  protected DialogHandler dhHandler = null;
  /**
   * SOAP-Namespace Identifier
   */
  String soapNSPrefix;
  /**
   * SOAP-Namespace Identifier
   */
  String osciNSPrefix;
  /**
   * SOAP-Namespace Identifier
   */
  String dsNSPrefix;
  /**
   * SOAP-Namespace Identifier
   */
  String xencNSPrefix;
  StringBuffer ns;
  OSCIMessageBuilder childBuilder = null;

  /**
   *  Konstruktor fr das OSCIMessageBuilder-Objekt
   *
   *@param  xmlReader
   *@param  dh
   */
  public OSCIEnvelopeBuilder(XMLReader xmlReader, DialogHandler dh)
  {
    this.xmlReader = xmlReader;
    this.dhHandler = dh;
    ns = new StringBuffer();
  }

  /**
   * undocumented
   *
   * @param prefix undocumented
   * @param uri undocumented
   */
  public void startPrefixMapping(String prefix, String uri)
  {
    if (uri.equals(OSCIMessageBuilder.SOAP_XMLNS))
      soapNSPrefix = prefix;
    else if (uri.equals(OSCIMessageBuilder.OSCI_XMLNS))
      osciNSPrefix = prefix;
    else if (uri.equals(OSCIMessageBuilder.DS_XMLNS))
      dsNSPrefix = prefix;
    else if (uri.equals(OSCIMessageBuilder.XENC_XMLNS))
      xencNSPrefix = prefix;

    ns.append(" xmlns:" + prefix + "=\"" + uri + "\"");
  }

  /**
   *  Start element.
   *
   *@param  uri                           Description of Parameter
   *@param  localName                     Description of Parameter
   *@param  qName                         Description of Parameter
   *@param  attributes                    Description of Parameter
   *@exception  SAXException  Description of Exception
   */
  public void startElement(String uri, String localName, String qName, Attributes attributes)
                    throws SAXException
  {
    if (log.isDebugEnabled())
      log.debug("Start-Element: " + localName + "\n" + uri + "\n" + qName + attributes.getLength());

    // SOAP-Namespacedefinition suchen
    if ((soapNSPrefix == null) || !localName.equals("Envelope"))
      throw new SAXException(DialogHandler.text.getString("msg_format_error") + ": " + localName);

    String schemaLocation = attributes.getValue("xsi:schemaLocation");

    if (log.isDebugEnabled())
      log.debug("Nachrichtentyp: " + schemaLocation);

    try
    {
      if (schemaLocation == null)
      {
        childBuilder = new SOAPFaultBuilder(this);
      }
      else if (schemaLocation.equals(XSD_ENCRYPTED_DATA))
      {
        try
        {
          childBuilder = new SOAPMessageEncryptedBuilder(this);
        }
        catch (Exception ex)
        {
          throw new SAXException(ex);
        }
      }
      else if (schemaLocation.equals(XSD_ACCEPT_DELIVERY))
      {
        childBuilder = new AcceptDeliveryBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_FORWARD_DELIVERY))
      {
        childBuilder = new ResponseToForwardDeliveryBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_INIT_DIALOG))
      {
        childBuilder = new ResponseToInitDialogBuilder(this);
      }
      else if (schemaLocation.equals(XSD_PROCESS_DELIVERY))
      {
        childBuilder = new ProcessDeliveryBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_MEDIATE_DELIVERY))
      {
        childBuilder = new ResponseToMediateDeliveryBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_GET_MSG_ID))
      {
        childBuilder = new ResponseToGetMessageIdBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_EXIT_DIALOG))
      {
        childBuilder = new ResponseToExitDialogBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_STORE_DELIVERY))
      {
        childBuilder = new ResponseToStoreDeliveryBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_FETCH_DELIVERY))
      {
        childBuilder = new ResponseToFetchDeliveryBuilder(this);
      }
      else if (schemaLocation.equals(XSD_RSP_FETCH_PROCESS_CARD))
      {
        childBuilder = new ResponseToFetchProcessCardBuilder(this);
      }
      else
      {
        log.error("Falsche OSCI-Nachricht. Nachrichtentyp nicht bekannt! Nachrichtentypt ist:\n" + schemaLocation +
                  "\n" + (XSD_RSP_FETCH_PROCESS_CARD));
        throw new SAXException(DialogHandler.text.getString("sax_exception_msgtype"));
      }

      this.xmlReader.setContentHandler(childBuilder);
      childBuilder.getOSCIMessage().soapNSPrefix = soapNSPrefix;
      childBuilder.getOSCIMessage().osciNSPrefix = osciNSPrefix;
      childBuilder.getOSCIMessage().dsNSPrefix = dsNSPrefix;
      childBuilder.getOSCIMessage().xencNSPrefix = xencNSPrefix;
      childBuilder.getOSCIMessage().ns = ns.toString();
    }
    catch (NoSuchAlgorithmException ex)
    {
      throw new SAXException(ex);
    }
    catch (OSCIRoleException ex)
    {
      throw new SAXException(ex);
    }
  }

  /**
   * undocumented
   */
  public void endDocument()
  {
    childBuilder.getOSCIMessage().hashableMsgPart = hashNCanStream.getDigestValues();

    if (childBuilder.getOSCIMessage().signatureHeader != null)
      childBuilder.getOSCIMessage().signatureHeader.signedInfo = (byte[]) hashNCanStream.getSignedInfos().remove(0);

    childBuilder.getOSCIMessage().stateOfMsg |= OSCIMessage.STATE_PARSED;
  }

  /**
   * undocumented
   *
   * @return undocumented
   */
  public XMLReader getXmlReader()
  {
    return xmlReader;
  }
}
