package de.osci.osci12.messagetypes;

import de.osci.helper.StoreOutputStream;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.roles.Addressee;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.soapheader.OsciH;

import java.io.IOException;
import java.io.OutputStream;

import java.security.NoSuchAlgorithmException;
import java.util.Vector;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte fr Annahmeantworten
 * angelegt. Ein passiver Client, der als Supplier fungiert, muss nach Empfang
 * eines Annahmeauftrags eine Instanz dieser Klasse aufbauen und an den Intermedir
 * zurcksenden. Die Nachricht enthlt inhaltlich lediglich eine Rckmeldung ber
 * den Empfang der Nachricht (Feedback).</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 *
 * @see de.osci.osci12.messagetypes.AcceptDelivery
 */
public class ResponseToAcceptDelivery extends OSCIResponseTo
{
  //  private static Log log = LogFactory.getLog(ResponseToAcceptDelivery.class);
  /**
   * Legt ein Nachrichtenobjekt fr eine Annahmeantwort an.
   *
   * @param request Auftragsnachricht
   */
  public ResponseToAcceptDelivery(AcceptDelivery request)
  {
    super(request.getDialogHandler());
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    // Hier wurde die Spec korrigiert
    dialogHandler.getControlblock().setChallenge(null);
    dialogHandler.getControlblock().setConversationID(null);
    dialogHandler.getControlblock().setSequenceNumber(-1);
    messageType = RESPONSE_TO_ACCEPT_DELIVERY;
  }

  ResponseToAcceptDelivery(DialogHandler dh)
  {
    super(dh);
    addressee = ((Addressee) dh.getSupplier());
    messageType = RESPONSE_TO_ACCEPT_DELIVERY;
  }

  /**
   * Setzt die Rckmeldungen (Fehler und Warnungen) auf Auftragsebene.
   * @param code Array mit Fehlercodes
   */
  public void setFeedback(String[] code)
  {
    super.setFeedback(code);
  }

  /**
   * Setzt die Rckmeldungen (Feedback-Eintrge) mit frei whlbaren Texten und Sprachen.
   * Der erste Index des Arrays entspricht dem Index des Entry-Elementes.
   * Beim zweiten Index bezeichnet <br>
   * 0 - das Sprachkrzel (z.B. "de", "en-US")<br>
   * 1 - den Code<br>
   * 2 - den Text<br>
   */
  public void setFeedback(String[][] entries)
  {
    feedBack = new Vector();
    for (int i = 0; i < entries.length; i++)
      feedBack.add(entries[i]);
  }

  /**
   * Bringt eine Supplier-Signatur an.
   * @throws IOException bei Schreib-/Leseproblemen
   * @throws OSCIRoleException wenn dem Rollenobjekt, das als Client fungiert,
   * kein Signer-Objekt zugeordnet wurde
   * @throws de.osci.osci12.common.OSCICancelledException bei Abbruch durch den
   * Benutzer
   */
  public void sign()
            throws IOException,
                   OSCIException,
                   de.osci.osci12.common.OSCICancelledException,
                   java.security.NoSuchAlgorithmException
  {
    super.sign(dialogHandler.getSupplier());
    //    messageParts.set(1, signatureHeader);
    messageParts.set(2, nonIntermediaryCertificatesH);
  }

  /**
   * undocumented
   *
   * @throws OSCIRoleException undocumented
   */
   protected void compose() throws OSCIException,
                                  NoSuchAlgorithmException,
                               IOException
  {
    super.compose();
    osciH = new OsciH("responseToAcceptDelivery", writeFeedBack());
    messageParts.add(osciH);
    messageParts.addAll(customHeaders);
    body = new Body("");
    body.setNSPrefixes(this);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }

  /**
   * Serialisiert und schreibt die Nachricht - ggf. verschlsselt - in den
   * bergebenen Stream. Es wird abhngig vom isEncryption()-Flag des
   * DialogHandlers verschlsselt oder nicht, Signaturen mssen jedoch von der
   * Anwendung selbst vor Aufruf dieser Methode (sign()) angebracht werden.
   * Die ausgehende Nachricht kann zu Debug- oder Archivierungszwecken (in jedem Fall
   * unverschlsselt) in den zweiten bergebenen Stream geschrieben werden.
   * Dieser Parameter kann null sein.
   * @param out Stream, in den die Antwortnachricht geschrieben werden soll
   * @param storeOutput Stream, in dem die (unverschlsselte) Antwortnachricht
   * gespeichert werden soll
   * @throws OSCIRoleException wenn erforderliche Zertifikate fehlen
   * @throws IOException bei Schreibproblemen
   * @throws java.security.NoSuchAlgorithmException wenn ein bentigter
   * Algorithmus nicht untersttzt wird
   * @see #sign()
   */
  public void writeToStream(OutputStream out, OutputStream storeOutput)
                          throws IOException,
                            de.osci.osci12.OSCIException,
                       NoSuchAlgorithmException
  {
    if (dialogHandler.isEncryption())
      new SOAPMessageEncrypted(this, storeOutput).writeXML(out);
    else if (storeOutput != null)
    {
      StoreOutputStream sos = new StoreOutputStream(out, storeOutput);
      writeXML(sos);
      sos.close();
    }
    else
      writeXML(out);
  }
}
