package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Attachment;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.Content;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.messageparts.Inspection;
import de.osci.osci12.messageparts.MessagePartsFactory;
import de.osci.osci12.messageparts.ProcessCardBundle;
import de.osci.osci12.messageparts.Timestamp;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Dieses Klasse reprsentiert die Antwort des Intermedirs auf einen
 * Zustellungsabholauftrag.
 * Clients erhalten vom Intermedir eine Instanz dieser Klasse, die eine Rckmeldung
 * ber den Erfolg der Operation (getFeedback()) sowie ggf. die angeforderten verschlsselten
 * und/oder unverschlsselten Inhaltsdaten einschl. des zugehrigen Laufzettels
 * enthlt.</p>
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 * @see de.osci.osci12.messagetypes.FetchDelivery
 */
public class ResponseToFetchDelivery extends OSCIResponseTo implements ContentPackageI
{
  //  private static Log log = LogFactory.getLog(ResponseToFetchDelivery.class);
  private String selectionRule;
  private int selectionMode = -1;
  public ProcessCardBundle processCardBundle;

  ResponseToFetchDelivery(FetchDelivery fetchDel, StoreDelivery storeDel)
                   throws NoSuchAlgorithmException, 
                          OSCIRoleException
  {
    super(fetchDel.dialogHandler);
    //hier werden die Rollenobjekte getauscht...
    messageType = RESPONSE_TO_FETCH_DELIVERY;
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    setSelectionMode(fetchDel.getSelectionMode());
    setSelectionRule(fetchDel.getSelectionRule());

    if (storeDel == null)
    {
      setFeedback(new String[] { "9803" });
    }
    else
    {
      addressee = storeDel.addressee;
      originator = storeDel.originator;

      ContentContainer[] con = storeDel.getContentContainer();

      for (int i = 0; i < con.length; i++)
        addContentContainer(con[i]);

      EncryptedDataOSCI[] enc = storeDel.getEncryptedData();

      for (int i = 0; i < enc.length; i++)
        addEncryptedData(enc[i]);

      Attachment[] att = storeDel.getAttachments();

      for (int i = 0; i < att.length; i++)
        addAttachment(att[i]);

      for (int i = 0; i < storeDel.getOtherAuthors().length; i++)
        otherAutors.put(storeDel.getOtherAuthors()[i].id, storeDel.getOtherAuthors()[i]);

      for (int i = 0; i < storeDel.getOtherReaders().length; i++)
        otherReaders.put(storeDel.getOtherReaders()[i].id, storeDel.getOtherReaders()[i]);
    }
  }

  ResponseToFetchDelivery(DialogHandler dh)
  {
    super(dh);
    originator = ((Originator) dh.getClient());
    messageType = RESPONSE_TO_FETCH_DELIVERY;
  }

  /**
   * Liefert die in die Nachricht eingestellten (unverschlsselten) Inhaltsdaten als ContentContainer-Objekte.
   * @return enthaltene ContentContainer mit Inhaltsdaten
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> Inhaltsdaten nach dem ContentContainer
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen ContentContainer oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public ContentContainer getContentContainerByRefID(String refID)
  {
	  return super.getContentContainerByRefID(refID);
  }
  
  /**
   * Durchsucht <b>die unverschlsselten</b> ContentContainer nach dem Content
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen Content oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public Content getContentByRefID(String refID)
  {
	return super.getContentByRefID(refID);  
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten
   * als EncryptedData-Objekte.
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  void setSelectionRule(String selectionRule)
  {
    this.selectionRule = selectionRule;
  }

  /**
   * Liefert die gesetzte Auswahlregel. Der Inhalt des zurckgegebenen
   * Strings hngt vom gesetzten Auswahlmodus ab und kann
   * entweder in einer Base64-codierten Message-Id oder einem Datum bestehen.
   * Das Format eines Datums entspricht dem XML-Schema nach
   * <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">
   * http://www.w3.org/TR/xmlschema-2/#dateTime</a>.
   * Wurde keine Regel gesetzt,
   * wird als default null zurckgegeben.
   * @return Auswahlregel (Message-Id oder Datum)
   * @see #getSelectionRule()
   */
  public String getSelectionRule()
  {
    return selectionRule;
  }

  void setSelectionMode(int selectionMode)
  {
    this.selectionMode = selectionMode;
  }

  /**
   * Liefert den gesetzten Auswahlmodus.
   * @return den Auswahlmodus SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION
   * oder NO_SELECTION_RULE
   */
  public int getSelectionMode()
  {
    return selectionMode;
  }

  /**
   * Diese Methode liefert den Laufzettel der Zustellung zurck oder null,
   * wenn bei der Verarbeitung der Nachricht ein Fehler aufgetreten ist.
   * Die Informationen im Laufzettel knnen auch direkt ber die einzelnen
   * getX()-Methoden ausgewertet werden.
   *
   * @return den Laufzettel als ProcessCardBundle-Objekt, im Fehlerfall null
   * @see #getTimestampCreation()
   * @see #getTimestampForwarding()
   * @see #getInspections()
   * @see #getSubject()
   * @see #getRecentModification()
   * @see #getMessageId()
   */
  public ProcessCardBundle getProcessCardBundle()
  {
    return processCardBundle;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des Eingangs
   * des Zustellungsauftrags beim Intermedir.
   * @return Zeitstempel der Einreichung beim Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampCreation()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getCreation();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des
   * vollstndigen Aufbaus der Abholantwort vom Intermedir fr den Empfnger.
   * @return Zeitstempel der Erstellung der Abholantwort durch den Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampForwarding()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getForwarding();
  }

  /**
   * Liefert die Ergebnisse der Zertifikatsprfungen in Form von Inspection-Objekten,
   * die im ProcessCardBundle-Objekt enthalten sind.
   * @return inspections die Prfergebnisse
   * @see #getProcessCardBundle()
   */
  public Inspection[] getInspections()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getInspections();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   * @see #getProcessCardBundle()
   */
  public String getSubject()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getSubject();
  }

  /**
   * Liefert das Datum der letzten nderung des Laufzettels. Das Format
   * entspricht dem XML-Schema nach http://www.w3.org/TR/xmlschema-2/#dateTime
   * @return Datum der letzten nderung
   * @see #getProcessCardBundle()
   */
  public String getRecentModification()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getRecentModification();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   */
  public String getMessageId()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getMessageId();
  }

  void sign()
     throws IOException, 
            OSCIException, 
            de.osci.osci12.common.OSCICancelledException, 
            java.security.NoSuchAlgorithmException
  {
    super.sign();
    // Der IntemedCertHeader liegt hier weiter unten
    messageParts.set(4, intermediaryCertificatesH);
    messageParts.set(2, null);
  }

  /**
   * undocumented
   *
   * @throws OSCIException undocumented
   * @throws IOException undocumented
   */
  protected void compose() throws OSCIException, 
                                  IOException, 
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(2, null);
    createNonIntermediaryCertificatesH();

    String selection = "";

    if (selectionMode == SELECT_BY_MESSAGE_ID)
    {
      try
      {
        selection = "<" + osciNSPrefix + ":fetchDelivery><" + osciNSPrefix + ":SelectionRule><" + osciNSPrefix +
                    ":MessageId>" + de.osci.helper.Base64.encode(selectionRule.getBytes(Constants.CHAR_ENCODING)) +
                    "</" + osciNSPrefix + ":MessageId></" + osciNSPrefix + ":SelectionRule></" + osciNSPrefix +
                    ":fetchDelivery>";
      }
      catch (UnsupportedEncodingException ex)
      {
      }
    }
    else if (selectionMode == SELECT_BY_DATE_OF_RECEPTION)
      selection = "<" + osciNSPrefix + ":fetchDelivery><" + osciNSPrefix + ":SelectionRule><" + osciNSPrefix +
                  ":ReceptionOfDelivery>" + selectionRule + "</" + osciNSPrefix + ":ReceptionOfDelivery></" +
                  osciNSPrefix + ":SelectionRule></" + osciNSPrefix + ":fetchDelivery>";
    else
      selection = "<" + osciNSPrefix + ":fetchDelivery></" + osciNSPrefix + ":fetchDelivery>";

    if (processCardBundle == null)
      osciH = new OsciH("responseToFetchDelivery", writeFeedBack() + selection);
    else
    {
      ByteArrayOutputStream out = new ByteArrayOutputStream();
      MessagePartsFactory.writeXML(processCardBundle, out);
      osciH = new OsciH("responseToFetchDelivery", writeFeedBack() + selection + out.toString(Constants.CHAR_ENCODING));
    }

    messageParts.add(osciH);
    messageParts.add(intermediaryCertificatesH);
    messageParts.add(nonIntermediaryCertificatesH);
    messageParts.addAll(customHeaders);
    body = new Body(getContentContainer(), getEncryptedData());
    stateOfMsg |= STATE_COMPOSED;
    messageParts.add(body);
  }
}
