package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.Inspection;
import de.osci.osci12.messageparts.ProcessCardBundle;
import de.osci.osci12.messageparts.Timestamp;
import de.osci.osci12.soapheader.OsciH;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.IOException;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Dieses Klasse reprsentiert die Antwort des Intermedirs auf einen
 * Weiterleitungsauftrag.
 * Clients erhalten vom Intermedir eine Instanz dieser Klasse, die eine Rckmeldung
 * ber den Erfolg der Operation (getFeedback()) sowie ggf. den zugehrigen
 * Laufzettel enthlt.</p>
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 * @see de.osci.osci12.messagetypes.ForwardDelivery
 */
public class ResponseToForwardDelivery extends OSCIResponseTo
{
  private static Log log = LogFactory.getLog(ResponseToForwardDelivery.class);
  ProcessCardBundle processCardBundle;

  /**
   * Dieser Konstruktor ist fr den Intermedir
   * @param accDel
   * @param rspAccDel
   */
  ResponseToForwardDelivery(AcceptDelivery accDel, ResponseToAcceptDelivery rspAccDel)
                     throws java.security.NoSuchAlgorithmException
  {
    super(accDel.fwdDH);
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      dialogHandler.getControlblock().setSequenceNumber(0);

    this.feedBack = rspAccDel.feedBack;
    body = new Body("");
    body.setNSPrefixes(this);
  }

  ResponseToForwardDelivery(DialogHandler dh)
  {
    super(dh);
    messageType = RESPONSE_TO_FORWARD_DELIVERY;
    body = new Body("");
    body.setNSPrefixes(this);
  }

  /**
   * Diese Methode liefert den Laufzettel der Zustellung zurck oder null,
   * wenn bei der Verarbeitung der Nachricht ein Fehler aufgetreten ist.
   * Die Informationen im Laufzettel knnen auch direkt ber die einzelnen
   * getX()-Methoden ausgewertet werden.
   *
   * @return den Laufzettel als ProcessCardBundle-Objekt, im Fehlerfall null
   * @see #getTimestampCreation()
   * @see #getTimestampForwarding()
   * @see #getTimestampReception()
   * @see #getInspections()
   * @see #getSubject()
   * @see #getRecentModification()
   * @see #getMessageId()
   */
  public ProcessCardBundle getProcessCardBundle()
  {
    return processCardBundle;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des Eingangs
   * des Weiterleitungsauftrags beim Intermedir.
   * @return Zeitstempel der Einreichung beim Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampCreation()
  {
    return processCardBundle.getCreation();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des
   * vollstndigen Aufbaus des Annahmeauftrags vom Intermedir fr den Empfnger.
   * @return Zeitstempel der Erstellung des Annahmeauftrags durch den Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampForwarding()
  {
    return processCardBundle.getForwarding();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des
   * Eingangs einer positiven Annahmeantwort vom Empfnger beim Intermedir.
   * @return Zeitstempel der Registrierung einer Empfangsbesttigung (Annahmeantwort)
   * durch den Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampReception()
  {
    return processCardBundle.getReception();
  }

  /**
   * Liefert die Ergebnisse der Zertifikatsprfungen in Form von Inspection-Objekten,
   * die im ProcessCardBundle-Objekt enthalten sind.
   * @return inspections die Prfergebnisse
   * @see #getProcessCardBundle()
   */
  public Inspection[] getInspections()
  {
    return processCardBundle.getInspections();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   * @see #getProcessCardBundle()
   */
  public String getSubject()
  {
    return processCardBundle.getSubject();
  }

  /**
   * Liefert das Datum der letzten nderung des Laufzettels. Das Format
   * entspricht dem XML-Schema nach http://www.w3.org/TR/xmlschema-2/#dateTime
   * @return Datum der letzten nderung
   * @see #getProcessCardBundle()
   */
  public String getRecentModification()
  {
    return processCardBundle.getRecentModification();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   * @see #getProcessCardBundle()
   */
  public String getMessageId()
  {
    return processCardBundle.getMessageId();
  }

  /**
   * undocumented
   *
   * @param processCardBundle undocumented
   */
  protected void setProcessCardBundle(ProcessCardBundle processCardBundle)
  {
    this.processCardBundle = processCardBundle;
  }

  void sign()
     throws IOException, 
            OSCIException, 
            de.osci.osci12.common.OSCICancelledException, 
            java.security.NoSuchAlgorithmException
  {
    super.sign();
    // Der NonIntemedCertHeader liegt hier weiter unten
    messageParts.set(4, intermediaryCertificatesH);
    messageParts.set(2, null);
  }

  /**
   * undocumented
   *
   * @throws OSCIException undocumented
   * @throws IOException undocumented
   * @throws NoSuchAlgorithmException undocumented
   */
  protected void compose() throws OSCIException, 
                                  IOException, 
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(2, null);

    if (processCardBundle == null)
      osciH = new OsciH("responseToForwardDelivery", writeFeedBack());
    else
      osciH = new OsciH("responseToForwardDelivery", writeFeedBack() + processCardBundle.writeToString());

    messageParts.add(osciH);
    messageParts.add(intermediaryCertificatesH);
    messageParts.addAll(customHeaders);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
