package de.osci.osci12.messagetypes;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.roles.Addressee;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;
import de.osci.osci12.soapheader.QualityOfTimestampH;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.IOException;
import java.io.OutputStream;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte fr Zustellungsauftrge
 * angelegt. Die Inhaltsdaten werden in Form von ContentContainer-Objekten
 * oder (im verschlsselten Fall) EncryptedData-Objekten in die Nachricht
 * eingestellt. Clients erhalten als Antwort auf diese Nachricht
 * vom Intermedir ein ResponseToStoreDelivery-Nachrichtenobjekt,
 * welches eine Rckmeldung ber den Erfolg der Operation (getFeedback())
 * und den Laufzettel der Zustellung enthlt.</p>
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 *
 * @see ResponseToStoreDelivery
 */
public class StoreDelivery extends OSCIRequest implements ContentPackageI
{
  protected static Log log = LogFactory.getLog(StoreDelivery.class);
  private String subject;

  StoreDelivery()
  {
    super();
    messageType = STORE_DELIVERY;
  }

  /**
   * Legt ein Nachrichtenobjekt fr einen Zustellungsauftrag an.
   *
   * @param dh DialogHandler-Objekt des Dialogs, innerhalb dessen die Nachricht
   * versendet werden soll
   * @param addressee Rollenobjekt des Empfngers
   * @param messageId Message-Id der Nachricht
   * @throws OSCIRoleException wenn das Verschlsselungszertifikat des Empfngers fehlt
   * @throws NoSuchAlgorithmException wenn der verwendete Security-Provider
   * einen erforderlichen Algorithmus nicht untersttzt (Erzeugung einer Zufallszahl)
   * @see de.osci.osci12.common.DialogHandler
   */
  public StoreDelivery(DialogHandler dh, Addressee addressee, String messageId)
                throws OSCIRoleException,
                       NoSuchAlgorithmException
  {
    super(dh);
    messageType = STORE_DELIVERY;
    originator = ((Originator) dh.getClient());
    this.addressee = addressee;
    // Check, ob ein Cipherzert eingestellt wurde
    addressee.getCipherCertificate();
    this.messageId = messageId;

    if (!dialogHandler.isExplicitDialog())
      dialogHandler.resetControlBlock();

    dialogHandler.getControlblock().setResponse(dialogHandler.getControlblock().getChallenge());
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock().setSequenceNumber(dialogHandler.getControlblock().getSequenceNumber() + 1);
    setQualityOfTimeStampCreation(false);
    setQualityOfTimeStampReception(false);
  }

  /**
   * Setzt die gewnschte Qualitt des Zeitstempels, mit dem der Intermedir
   * den Eingang des Auftrags im Laufzettel protokolliert.
   *
   * @param cryptographic <b>true</b>: kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b>: Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs, default)
   * @see #getQualityOfTimeStampCreation()
   */
  public void setQualityOfTimeStampCreation(boolean cryptographic)
  {
    qualityOfTimestampTypeCreation = new QualityOfTimestampH(false, cryptographic);
  }

  /**
   * Liefert die Qualitt des Zeitstempels, mit dem der Intermedir den
   * Eingang des Auftrags im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @see #setQualityOfTimeStampCreation(boolean)
   */
  public boolean getQualityOfTimeStampCreation()
  {
    return qualityOfTimestampTypeCreation.isQualityCryptographic();
  }

  /**
   * Setzt die gewnschte Qualitt des Zeitstempels, mit dem der Intermedir die
   * Empfangbesttigung der Zustellung durch den Empfnger im Laufzettel protokolliert.
   * Die Empfangsbesttigung besteht in einem weiteren Auftrag, den der Empfnger
   * nach Erhalt der Zustellungsabholantwort innerhalb desselben expliziten Dialogs
   * an den Intermedir schickt.
   * @param cryptographic <b>true</b>: kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b>: Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @see #getQualityOfTimeStampReception()
   */
  public void setQualityOfTimeStampReception(boolean cryptographic)
  {
    qualityOfTimestampTypeReception = new QualityOfTimestampH(true, cryptographic);
  }

  /**
   * Liefert die geforderte Qualitt des Zeitstempels, mit dem der Intermedir den
   * Empfang der Annahmeantwort im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs, default)
   * @see #setQualityOfTimeStampReception(boolean)
   */
  public boolean getQualityOfTimeStampReception()
  {
    return qualityOfTimestampTypeReception.isQualityCryptographic();
  }

  /**
   * Setzt den Betreff-Eintrag im Laufzettel. Der bergebene Parameter-String
   * mu den Vorschriften fr den primitiven XML-Datentyp "string" entsprechen,
   * darf also z.B. keine XML-Steuerzeichen (<, > o..) enthalten.
   * @param  subject  der Betreff
   */
  public void setSubject(String subject)
  {
    this.subject = subject;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   */
  public String getSubject()
  {
    return subject;
  }

  /**
   * Fgt der Nachricht einen Inhaltsdatencontainer hinzu. Diese Methode
   * sollte erst aufgerufen werden, wenn der Container vollstndig erstellt wurde.
   * <b>Hinweis: </b>ContentContainer-Objekte mit Attachments, die durch Entschlsselung
   * von EncryptedDataOSCI-Objekten gewonnen wurden, knnen hier nicht ohne weiteres
   * hinzugefgt werden. In diesem Fall mssen die Attachments zunchst ausgelesen und
   * die Daten in neu angelegten Attachment-Objekten wieder hinzugefgt werden.
   * @param container Inhaltsdatencontainer
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public void addContentContainer(ContentContainer container)
                           throws OSCIRoleException
  {
    super.addContentContainer(container);
  }

  /**
   * Entfernt einen Inhaltsdatencontainer aus der Nachricht.
   * @param container Inhaltsdatencontainer
   * @see #addContentContainer
   */
  public void removeContentContainer(ContentContainer container)
  {
    super.removeContentContainer(container);
  }

  /**
   * Liefert die in die Nachricht eingestellten (unverschlsselten) Inhaltsdaten als ContentContainer-Objekte.
   * @return enthaltene ContentContainer mit Inhaltsdaten
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Fgt der Nachricht einen EncryptedData-Eintrag mit verschlsselten
   * Inhaltsdaten hinzu.
   * @param encData verschlsselte Daten
   * @see EncryptedDataOSCI
   */
  public void addEncryptedData(EncryptedDataOSCI encData)
                        throws OSCIRoleException
  {
    super.addEncryptedData(encData);
  }

  /**
   * Entfernt einen EncryptedData-Eintrag mit verschlsselten Daten
   * aus der Nachricht.
   * @param encData verschlsselte Daten
   * @see #addEncryptedData
   * @see EncryptedDataOSCI
   */
  public void removeEncryptedData(EncryptedDataOSCI encData)
  {
    super.removeEncryptedData(encData);
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten
   * als EncryptedData-Objekte.
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   */
  public String getMessageId()
  {
    return messageId;
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Diese Methode wirft eine Exception, wenn beim Aufbau oder Versand der
   * Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen
   * dem Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send(OutputStream, OutputStream)
   */
  public ResponseToStoreDelivery send() throws IOException,
                                               de.osci.osci12.OSCIException,
                                               NoSuchAlgorithmException
  {
    return (ResponseToStoreDelivery) transmit(null, null);
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Die aus- und eingehenden Daten werden zustzlich in die bergebenen
   * Streams geschrieben (unverschlsselte Auftragsdaten). Diese Parameter
   * drfen null sein.
   *
   * @param storeOutput Stream, in den die versendete Nachricht geschrieben wird
   * @param storeInput Stream, in den die empfangene Antwortnachricht geschrieben wird
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send()
   */
  public ResponseToStoreDelivery send(OutputStream storeOutput, OutputStream storeInput)
                               throws IOException,
                                      de.osci.osci12.OSCIException,
                                      NoSuchAlgorithmException
  {
    return (ResponseToStoreDelivery) transmit(storeOutput, storeInput);
  }

  /**
   * Bringt eine Client-Signatur an.
   * @throws IOException bei Schreib-/Leseproblemen
   * @throws OSCIException wenn Zusammenstellen der Daten ein Problem auftritt.
   * @throws de.osci.osci12.common.OSCICancelledException bei Abbruch durch den
   * Benutzer
   */
  void sign()
     throws IOException,
            OSCIException,
            de.osci.osci12.common.OSCICancelledException,
            java.security.NoSuchAlgorithmException
  {
    super.sign();
    // Der NonIntemedCertHeader liegt hier weiter unten
    messageParts.set(3, null);
    messageParts.set(7, nonIntermediaryCertificatesH);
  }

  /**
   * undocumented
   *
   * @throws OSCIRoleException undocumented
   * @throws IOException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws OSCIException,
                                  IOException,
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(3, null);
    messageParts.add(qualityOfTimestampTypeCreation);
    messageParts.add(qualityOfTimestampTypeReception);

    if (messageId == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Message-Id");

    if (dialogHandler.getControlblock().getChallenge() == null)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": Challenge");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString("missing_entry") + ": SequenceNumber");

    importAllCertificates();

    String head = "<" + osciNSPrefix + ":MessageId>" +
                  de.osci.helper.Base64.encode(messageId.getBytes(Constants.CHAR_ENCODING)) + "</" + osciNSPrefix +
                  ":MessageId>";

    if (subject != null)
      head += ("<" + osciNSPrefix + ":Subject>" + subject + "</" + osciNSPrefix + ":Subject>");

    osciH = new OsciH("storeDelivery", head);
    messageParts.add(osciH);
    createNonIntermediaryCertificatesH();
    nonIntermediaryCertificatesH.getCipherCertificateAddressee(); // Test
    messageParts.add(nonIntermediaryCertificatesH);
    messageParts.addAll(customHeaders);
    body = new Body(getContentContainer(), getEncryptedData());
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
