package de.osci.osci12.messagetypes;

import de.osci.osci12.OSCIException;
import de.osci.osci12.messageparts.Attachment;
import de.osci.osci12.messageparts.Content;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.messageparts.ProcessCardBundle;
import de.osci.osci12.roles.Intermed;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import java.security.NoSuchAlgorithmException;


/**
 * <p>Eine Instanz dieser Klasse wird beim Einlesen einer serialisierten
 * OSCI-Nachricht beliebigen Typs (Nachrichten mit Inhaltsdaten) angelegt.</p>
 * Die Klasse dient folgenden Zwecken:<ul><li>Abspeichern und Wiedereinlesen von
 * Nachrichten mit Inhaltsdaten</li><li>Austausch von Inhaltsdaten zwischen
 * Autoren und Sendern bzw. Empfnger und Lesern.</li></ul>
 * <p>
 * Leser knnen ihre Inhaltsdaten in Nachrichten beliebigen Typs ablegen, diese
 * speichern und z.B. als Datei weiterreichen. Absender knnen eine solche Datei
 * mit der Methode loadMessage(InputStream) laden, die Inhaltdatencontainer
 * entnehmen und anderen Nachrichten hinzufgen.
 * </p><p>
 * Ein generelles Problem ist, dass in verschlsselten Inhaltsdaten einer
 * OSCI-Nachricht die Informationen ber die enthaltenen Referenzen auf
 * Zertifikate und Attachments ohne Entschlsselung nicht verfgbar sind.
 * Im Zweifel mssen daher alle Zertifikate (z.B. mit den Methoden
 * OSCIMessage.getOtherAuthors(), OSCIMessage.getOtherReaders() und
 * OSCIMessage.addRole(Role)) und Attachments (Methoden
 * exportAttachment(OSCIMessage, Attachment) und exportAttachments(OSCIMessage))
 * entnommen und der neuen Nachricht hinzugefgt werden.
 * </p><p>
 * Anwendungen sollten dies bercksichtigen und fr den Inhaltsdatenaustausch
 * mglichst mehrere einzelne Nachrichten statt einer komplexen verwenden.
 * Besonders problematisch ist in diesem Zusammenhang die Signatur durch
 * Originator- bzw. Verschlsselung fr Addressee-Rollenobjekte, weil diese
 * i.d.R. nicht in eine neue Nachricht bernommen werden knnen. Hier sollten
 * grundstzlich Author- und Reader-Objekte verwendet werden.</p>
 * <p>Weiter ist zu beachten, dass es beim Zusammensetzen einer neuen Nachricht aus
 * Inhaltdatencontainern, die anderen Nachrichten entnommen wurden, zu Konflikten
 * mit den Ref-Ids der Content-Eintrge kommen kann. Da die Bibliothek wegen der
 * ggf. vorhandenen Signatur diese Ids nicht selbst anpassen kann, sollten
 * Anwendungen eindeutige Ref-Ids setzen. Diese knnen z.B. aus Message-Ids
 * und laufenden Nummern oder Zertfikats-Ids (z.B. IssuerDN und SerialNumber)
 * und Datum/Uhrzeit generiert werden.</p>
 *
 *
 * <p>Copyright  2010 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.5
 */
public class StoredMessage extends OSCIResponseTo implements ContentPackageI
{
  //  private static Log log = LogFactory.getLog(StoredMessage.class);
  ProcessCardBundle[] processCardBundles;
  ProcessCardBundle processCardBundleReply;
  String subject;
  java.net.URI uriReceiver;
  String selectionRule;
  int selectionMode = -1;
  long quantityLimit = -1;

  StoredMessage(int messageType)
  {
    this.messageType = messageType;
  }

  /**
   * Serialisiert die bergebene Nachricht und schreibt die Daten in den bergebenen Stream.
   * @param msg zu speichernde Nachricht; z.Zt. knnen StoredMessage-Objekte selbst nicht
   * gespeichert werden. Der Stream wird von der Methode <b>nicht</b> geschlossen.
   * @param output Ausgabestream
   * @throws IOException bei Schreibfehlern
   * @throws OSCIException bei Problemen beim Aufbau der Nachricht
   * @throws NoSuchAlgorithmException wenn der installierte Security-Provider einen
   * bentigten Algorithmus nicht untersttzt
   * @see #loadMessage(InputStream)
   */
  public static void storeMessage(OSCIMessage msg, OutputStream output)
                           throws IOException,
                                  OSCIException,
                                  NoSuchAlgorithmException
  {
    if (msg instanceof StoredMessage)
      throw new UnsupportedOperationException();

    msg.writeXML(output);
  }

  /**
   * Liest eine (unverschlsselte) Nachricht aus dem bergebenen Stream. Die in
   * der Nachricht enthaltenen Inhalte knnen dem zurckgegebenen
   * Nachrichtenobjekt entnommen werden. <b>Die Methode prft keine Signaturen.</b>
   * @param input zu lesender Stream
   * @return StoredMessage-Instanz
   * @throws IOException bei Schreibfehlern
   * @throws OSCIException bei Problemen beim Aufbau der Nachricht
   * @throws NoSuchAlgorithmException wenn der installierte Security-Provider einen
   * bentigten Algorithmus nicht untersttzt
   * @see #storeMessage(OSCIMessage, OutputStream)
   * @see #loadMessageCheckingSignature(InputStream)
   */
  public static StoredMessage loadMessage(InputStream input)
                                   throws IOException,
                                          OSCIException,
                                          NoSuchAlgorithmException
  {
    return new StoredMessageParser().parseStream(input);
  }

  /**
   * Liest eine (unverschlsselte) Nachricht aus dem bergebenen Stream. Die in
   * der Nachricht enthaltenen Inhalte knnen dem zurckgegebenen
   * Nachrichtenobjekt entnommen werden. Diese Methode prft auerdem
   * eine ggf. vorhandene Nachrichtensignatur. Liefert die Prfung ein
   * negatives Ergebnis, wird eine OSCIException (Code 9601) geworfen.
   * <b>Die Methode prft keine Inhaltsdatensignaturen. Ebenso wird
   * keine Exception ausgelst, wenn die Nachricht unsigniert ist.</b>
   * @param input zu lesender Stream
   * @return StoredMessage-Instanz
   * @throws IOException bei Schreibfehlern
   * @throws OSCIException bei Problemen beim Aufbau der Nachricht oder ungltiger Nachrichtensignatur
   * @throws NoSuchAlgorithmException wenn der installierte Security-Provider einen
   * bentigten Algorithmus nicht untersttzt
   * @see #storeMessage(OSCIMessage, OutputStream)
   * @see #loadMessage(InputStream)
   */
  public static StoredMessage loadMessageCheckingSignature(InputStream input)
                                                    throws IOException,
                                                           OSCIException,
                                                           NoSuchAlgorithmException
  {
    StoredMessageParser smp = new StoredMessageParser();
    StoredMessage sm = smp.parseStream(input);

    if (sm.isSigned())
    {
      if (sm.getMessageType() <= 0x10)
        sm.dialogHandler.supplier.setSignatureCertificate(sm.dialogHandler.client.getSignatureCertificate());

      if (!smp.checkMsgHashes(sm))
        throw new OSCIException("9601");
    }

    return sm;
  }

  /**
   * Diese Methode exportiert alle Attachments der Nachricht in eine andere
   * OSCI-Nachricht beliebigen Typs.
   * @param destinationMessage Zielnachricht
   * @see #exportAttachment(OSCIMessage, Attachment)
   */
  public void exportAttachments(OSCIMessage destinationMessage)
  {
    Attachment[] atts = getAttachments();

    for (int i = 0; i < atts.length; i++)
      destinationMessage.addAttachment(atts[i]);
  }

  /**
   * Diese Methode exportiert ein Attachment der Nachricht in eine andere
   * OSCI-Nachricht beliebigen Typs.
   * @param destinationMessage Zielnachricht
   * @param att zu exportierendes Attachment
   * @see #exportAttachments(OSCIMessage)
   */
  public void exportAttachment(OSCIMessage destinationMessage, Attachment att)
  {
    Attachment[] atts = getAttachments();

    for (int i = 0; i < atts.length; i++)
    {
      if (att.equals(atts[i]))
      {
        destinationMessage.addAttachment(atts[i]);

        return;
      }
    }
  }

  /**
   * Diese Methode liefert den Laufzettel der Nachricht zurck.
   * Im Falle einer Abwicklungsantwort wird der Laufzettel des Requests
   * zurckgegeben.
   *
   * @return den Laufzettel des Auftrags als ProcessCardBundle-Objekt oder null,
   * wenn der Nachrichtentyp keinen Laufzettel enthlt
   * @see #getProcessCardBundleReply()
   * @see de.osci.osci12.messagetypes.ResponseToMediateDelivery#getProcessCardBundleRequest()
   */
  public ProcessCardBundle getProcessCardBundle()
  {
    if (processCardBundles == null)
      return null;

    return processCardBundles[0];
  }

  /**
   * Diese Methode liefert die Laufzettel einer Laufzettelabholantwort zurck.
   *
   * @return den Laufzettel des Auftrags als ProcessCardBundle-Objekt
   * @throws UnsupportedOperationException wenn die Nachricht nicht vom Typ
   * ResponseToFetchProcessCard ist
   * @see #getProcessCardBundle()
   * @see de.osci.osci12.messagetypes.ResponseToFetchProcessCard#getProcessCardBundles()
   */
  public ProcessCardBundle[] getProcessCardBundles()
  {
    if (messageType != OSCIMessage.RESPONSE_TO_FETCH_PROCESS_CARD)
      throw new UnsupportedOperationException();

    return processCardBundles;
  }

  /**
   * Diese Methode liefert den Antwort-Laufzettel einer Abwicklungsantwort zurck.
   *
   * @return den Laufzettel der Nachricht als ProcessCardBundle-Objekt
   * @throws UnsupportedOperationException wenn die Nachricht nicht vom Typ
   * ResponseToMediateDelivery ist.
   * @see #getProcessCardBundle()
   */
  public ProcessCardBundle getProcessCardBundleReply()
  {
    if (messageType != OSCIMessage.RESPONSE_TO_MEDIATE_DELIVERY)
      throw new UnsupportedOperationException();

    return processCardBundleReply;
  }

  /**
   * Liefert die Message-Id der Nachricht.
   * Im Falle einer Abwicklungsantwort wird die Message-Id des Requests
   * zurckgegeben.
   * @return Message-ID
   */
  public String getMessageId()
  {
    return messageId;
  }

  void setMessageId(String messageId)
  {
    this.messageId = messageId;
  }

  /**
   * Liefert das Intermedirsobjekt oder null, wenn keine Zertifikate in der
   * Nachricht enthalten sind.
   * @return Intermedir
   */
  public Intermed getIntermediary()
  {
    if (dialogHandler.getSupplier() instanceof Intermed)
      return (Intermed) dialogHandler.getSupplier();
    else

      return (Intermed) dialogHandler.getClient();
  }

  /**
   * Liefert die in die Nachricht eingestellten (unverschlsselten) Inhaltsdaten als ContentContainer-Objekte.
   * @return enthaltene ContentContainer mit Inhaltsdaten
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten
   * als EncryptedData-Objekte.
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> Inhaltsdaten nach dem ContentContainer
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen ContentContainer oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public ContentContainer getContentContainerByRefID(String refID)
  {
      return super.getContentContainerByRefID(refID);
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> ContentContainer nach dem Content
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen Content oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public Content getContentByRefID(String refID)
  {
    return super.getContentByRefID(refID);
  }

  /**
   * Liefert den Betreff der Nachricht oder null, wenn kein Betreff in der
   * Nachricht enthalten ist.
   * @return den Betreff
   */
  public String getSubject()
  {
    return subject;
  }

  /**
   * Liefert die URI des Nachrichtenempfngers oder null, wenn keine Empfnger-URI
   * in der Nachricht enthalten ist.
   * @return URI
   */
  public java.net.URI getUriReceiver()
  {
    return uriReceiver;
  }

  /**
   * Liefert die Qualitt des Zeitstempels, mit dem der Intermedir den
   * Eingang des Auftrags im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @throws UnsupportedOperationException wenn der reprsentierte Nachrichtentyp
   * keine Zeitstempelanforderung enthlt
   * @see #getQualityOfTimeStampReception()
   */
  public boolean getQualityOfTimeStampCreation()
  {
    if ((messageType != OSCIMessage.STORE_DELIVERY) && (messageType != OSCIMessage.FORWARD_DELIVERY) &&
            (messageType != OSCIMessage.MEDIATE_DELIVERY) && (messageType != OSCIMessage.RESPONSE_TO_PROCESS_DELIVERY))
      throw new UnsupportedOperationException();

    return qualityOfTimestampTypeCreation.isQualityCryptographic();
  }

  /**
   * Liefert die geforderte Qualitt des Zeitstempels, mit dem der Intermedir den
   * Empfang der Annahmeantwort im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @throws UnsupportedOperationException wenn der reprsentierte Nachrichtentyp
   * keine Zeitstempelanforderung enthlt
   * @see #getQualityOfTimeStampCreation()
   */
  public boolean getQualityOfTimeStampReception()
  {
    if ((messageType != OSCIMessage.STORE_DELIVERY) && (messageType != OSCIMessage.FORWARD_DELIVERY) &&
            (messageType != OSCIMessage.MEDIATE_DELIVERY) && (messageType != OSCIMessage.RESPONSE_TO_PROCESS_DELIVERY))
      throw new UnsupportedOperationException();

    return qualityOfTimestampTypeReception.isQualityCryptographic();
  }

  /**
   * Liefert den gesetzten Auswahlmodus fr Nachrichten oder Laufzettel.
   * @return den Auswahlmodus SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION
   * oder NO_SELECTION_RULE
   * @throws UnsupportedOperationException wenn der reprsentierte Nachrichtentyp
   * keine Auswahlmglichkeit enthlt
   * @see #getSelectionRule()
   */
  public int getSelectionMode()
  {
    if ((messageType != OSCIMessage.FETCH_DELIVERY) && (messageType != OSCIMessage.FETCH_PROCESS_CARD) &&
            (messageType != OSCIMessage.RESPONSE_TO_FETCH_DELIVERY) &&
            (messageType != OSCIMessage.RESPONSE_TO_FETCH_PROCESS_CARD))
      throw new UnsupportedOperationException();

    return selectionMode;
  }

  /**
   * Liefert die gesetzte Auswahlregel fr Nachrichten oder Laufzettel.
   * Wurde keine Regel gesetzt, wird als default null zurckgegeben.
   * @return Auswahlregel (Message-Id oder Datum)
   * @throws UnsupportedOperationException wenn der reprsentierte Nachrichtentyp
   * keine Auswahlmglichkeit enthlt
   * @see #getSelectionMode()
   */
  public String getSelectionRule()
  {
    if ((messageType != OSCIMessage.FETCH_DELIVERY) && (messageType != OSCIMessage.FETCH_PROCESS_CARD) &&
            (messageType != OSCIMessage.RESPONSE_TO_FETCH_DELIVERY) &&
            (messageType != OSCIMessage.RESPONSE_TO_FETCH_PROCESS_CARD))
      throw new UnsupportedOperationException();

    return selectionRule;
  }

  /**
   * Liefert die maximale Anzahl zurckzugebender Laufzettel.
   * @return gesetzte maximale Anzahl
   * @throws UnsupportedOperationException wenn die reprsentierte Nachricht
   * kein Laufzettelabholauftrag oder -antwort ist.
   */
  public long getQuantityLimit()
  {
    if ((messageType != OSCIMessage.FETCH_PROCESS_CARD) && (messageType != OSCIMessage.RESPONSE_TO_FETCH_PROCESS_CARD))
      throw new UnsupportedOperationException();

    return quantityLimit;
  }
}
